"""Test the PersistentPg module.

Note:
We don't test performance here, so the test does not predicate
whether PersistentPg actually will help in improving performance or not.
We also assume that the underlying SteadyPg connections are tested.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_pg as pg

from dbutils.persistent_pg import PersistentPg


class TestPersistentPg(unittest.TestCase):

    def test_version(self):
        from dbutils import __version__, persistent_pg
        self.assertEqual(persistent_pg.__version__, __version__)
        self.assertEqual(PersistentPg.version, __version__)

    def test_close(self):
        for closeable in (False, True):
            persist = PersistentPg(closeable=closeable)
            db = persist.connection()
            self.assertTrue(db._con.db and db._con.valid)
            db.close()
            self.assertTrue(
                closeable ^ (db._con.db is not None and db._con.valid))
            db.close()
            self.assertTrue(
                closeable ^ (db._con.db is not None and db._con.valid))
            db._close()
            self.assertFalse(db._con.db and db._con.valid)
            db._close()
            self.assertFalse(db._con.db and db._con.valid)

    def test_threads(self):
        num_threads = 3
        persist = PersistentPg()
        from queue import Queue, Empty
        query_queue, result_queue = [], []
        for i in range(num_threads):
            query_queue.append(Queue(1))
            result_queue.append(Queue(1))

        def run_queries(i):
            this_db = persist.connection().db
            while 1:
                try:
                    try:
                        q = query_queue[i].get(1, 1)
                    except TypeError:
                        q = query_queue[i].get(1)
                except Empty:
                    q = None
                if not q:
                    break
                db = persist.connection()
                if db.db != this_db:
                    r = 'error - not persistent'
                else:
                    if q == 'ping':
                        r = 'ok - thread alive'
                    elif q == 'close':
                        db.db.close()
                        r = 'ok - connection closed'
                    else:
                        r = db.query(q)
                r = f'{i}({db._usage}): {r}'
                try:
                    result_queue[i].put(r, 1, 1)
                except TypeError:
                    result_queue[i].put(r, 1)
            db.close()

        from threading import Thread
        threads = []
        for i in range(num_threads):
            thread = Thread(target=run_queries, args=(i,))
            threads.append(thread)
            thread.start()
        for i in range(num_threads):
            try:
                query_queue[i].put('ping', 1, 1)
            except TypeError:
                query_queue[i].put('ping', 1)
        for i in range(num_threads):
            try:
                r = result_queue[i].get(1, 1)
            except TypeError:
                r = result_queue[i].get(1)
            self.assertEqual(r, f'{i}(0): ok - thread alive')
            self.assertTrue(threads[i].is_alive())
        for i in range(num_threads):
            for j in range(i + 1):
                try:
                    query_queue[i].put(f'select test{j}', 1, 1)
                    r = result_queue[i].get(1, 1)
                except TypeError:
                    query_queue[i].put(f'select test{j}', 1)
                    r = result_queue[i].get(1)
                self.assertEqual(r, f'{i}({j + 1}): test{j}')
        try:
            query_queue[1].put('select test4', 1, 1)
            r = result_queue[1].get(1, 1)
        except TypeError:
            query_queue[1].put('select test4', 1)
            r = result_queue[1].get(1)
        self.assertEqual(r, '1(3): test4')
        try:
            query_queue[1].put('close', 1, 1)
            r = result_queue[1].get(1, 1)
        except TypeError:
            query_queue[1].put('close', 1)
            r = result_queue[1].get(1)
        self.assertEqual(r, '1(3): ok - connection closed')
        for j in range(2):
            try:
                query_queue[1].put(f'select test{j}', 1, 1)
                r = result_queue[1].get(1, 1)
            except TypeError:
                query_queue[1].put(f'select test{j}', 1)
                r = result_queue[1].get(1)
            self.assertEqual(r, f'1({j + 1}): test{j}')
        for i in range(num_threads):
            self.assertTrue(threads[i].is_alive())
            try:
                query_queue[i].put('ping', 1, 1)
            except TypeError:
                query_queue[i].put('ping', 1)
        for i in range(num_threads):
            try:
                r = result_queue[i].get(1, 1)
            except TypeError:
                r = result_queue[i].get(1)
            self.assertEqual(r, f'{i}({i + 1}): ok - thread alive')
            self.assertTrue(threads[i].is_alive())
        for i in range(num_threads):
            try:
                query_queue[i].put(None, 1, 1)
            except TypeError:
                query_queue[i].put(None, 1)

    def test_maxusage(self):
        persist = PersistentPg(20)
        db = persist.connection()
        self.assertEqual(db._maxusage, 20)
        for i in range(100):
            r = db.query(f'select test{i}')
            self.assertEqual(r, f'test{i}')
            self.assertTrue(db.db.status)
            j = i % 20 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db.num_queries, j)

    def test_setsession(self):
        persist = PersistentPg(3, ('set datestyle',))
        db = persist.connection()
        self.assertEqual(db._maxusage, 3)
        self.assertEqual(db._setsession_sql, ('set datestyle',))
        self.assertEqual(db.db.session, ['datestyle'])
        db.query('set test')
        for i in range(3):
            self.assertEqual(db.db.session, ['datestyle', 'test'])
            db.query('select test')
        self.assertEqual(db.db.session, ['datestyle'])

    def test_failed_transaction(self):
        persist = PersistentPg()
        db = persist.connection()
        db._con.close()
        self.assertEqual(db.query('select test'), 'test')
        db.begin()
        db._con.close()
        self.assertRaises(pg.InternalError, db.query, 'select test')
        self.assertEqual(db.query('select test'), 'test')
        db.begin()
        self.assertEqual(db.query('select test'), 'test')
        db.rollback()
        db._con.close()
        self.assertEqual(db.query('select test'), 'test')

    def test_context_manager(self):
        persist = PersistentPg()
        with persist.connection() as db:
            db.query('select test')
            self.assertEqual(db.num_queries, 1)


if __name__ == '__main__':
    unittest.main()



"""Test the SteadyDB module.

Note:
We do not test any real DB-API 2 module, but we just
mock the basic DB-API 2 connection functionality.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_db as dbapi

from dbutils.steady_db import (
    connect as SteadyDBconnect, SteadyDBConnection, SteadyDBCursor)


class TestSteadyDB(unittest.TestCase):

    def test_version(self):
        from dbutils import __version__, steady_db
        self.assertEqual(steady_db.__version__, __version__)
        self.assertEqual(steady_db.SteadyDBConnection.version, __version__)

    def test_mocked_connection(self):
        db = dbapi.connect(
            'SteadyDBTestDB', user='SteadyDBTestUser')
        db.__class__.has_ping = False
        db.__class__.num_pings = 0
        self.assertTrue(hasattr(db, 'database'))
        self.assertEqual(db.database, 'SteadyDBTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'SteadyDBTestUser')
        self.assertTrue(hasattr(db, 'cursor'))
        self.assertTrue(hasattr(db, 'close'))
        self.assertTrue(hasattr(db, 'open_cursors'))
        self.assertTrue(hasattr(db, 'num_uses'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertTrue(hasattr(db, 'session'))
        self.assertTrue(hasattr(db, 'valid'))
        self.assertTrue(db.valid)
        self.assertEqual(db.open_cursors, 0)
        for i in range(3):
            cursor = db.cursor()
            self.assertEqual(db.open_cursors, 1)
            cursor.close()
            self.assertEqual(db.open_cursors, 0)
        cursor = []
        for i in range(3):
            cursor.append(db.cursor())
            self.assertEqual(db.open_cursors, i + 1)
        del cursor
        self.assertEqual(db.open_cursors, 0)
        cursor = db.cursor()
        self.assertTrue(hasattr(cursor, 'execute'))
        self.assertTrue(hasattr(cursor, 'fetchone'))
        self.assertTrue(hasattr(cursor, 'callproc'))
        self.assertTrue(hasattr(cursor, 'close'))
        self.assertTrue(hasattr(cursor, 'valid'))
        self.assertTrue(cursor.valid)
        self.assertEqual(db.open_cursors, 1)
        for i in range(3):
            self.assertEqual(db.num_uses, i)
            self.assertEqual(db.num_queries, i)
            cursor.execute(f'select test{i}')
            self.assertEqual(cursor.fetchone(), f'test{i}')
        self.assertTrue(cursor.valid)
        self.assertEqual(db.open_cursors, 1)
        for i in range(4):
            cursor.callproc('test')
        cursor.close()
        self.assertFalse(cursor.valid)
        self.assertEqual(db.open_cursors, 0)
        self.assertEqual(db.num_uses, 7)
        self.assertEqual(db.num_queries, 3)
        self.assertRaises(dbapi.InternalError, cursor.close)
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test')
        self.assertTrue(db.valid)
        self.assertFalse(db.__class__.has_ping)
        self.assertEqual(db.__class__.num_pings, 0)
        self.assertRaises(AttributeError, db.ping)
        self.assertEqual(db.__class__.num_pings, 1)
        db.__class__.has_ping = True
        self.assertIsNone(db.ping())
        self.assertEqual(db.__class__.num_pings, 2)
        db.close()
        self.assertFalse(db.valid)
        self.assertEqual(db.num_uses, 0)
        self.assertEqual(db.num_queries, 0)
        self.assertRaises(dbapi.InternalError, db.close)
        self.assertRaises(dbapi.InternalError, db.cursor)
        self.assertRaises(dbapi.OperationalError, db.ping)
        self.assertEqual(db.__class__.num_pings, 3)
        db.__class__.has_ping = False
        db.__class__.num_pings = 0

    def test_broken_connection(self):
        self.assertRaises(TypeError, SteadyDBConnection, None)
        self.assertRaises(TypeError, SteadyDBCursor, None)
        db = SteadyDBconnect(dbapi, database='ok')
        for i in range(3):
            db.close()
        del db
        self.assertRaises(
            dbapi.OperationalError, SteadyDBconnect, dbapi, database='error')
        db = SteadyDBconnect(dbapi, database='ok')
        cursor = db.cursor()
        for i in range(3):
            cursor.close()
        cursor = db.cursor('ok')
        for i in range(3):
            cursor.close()
        self.assertRaises(dbapi.OperationalError, db.cursor, 'error')

    def test_close(self):
        for closeable in (False, True):
            db = SteadyDBconnect(dbapi, closeable=closeable)
            self.assertTrue(db._con.valid)
            db.close()
            self.assertTrue(closeable ^ db._con.valid)
            db.close()
            self.assertTrue(closeable ^ db._con.valid)
            db._close()
            self.assertFalse(db._con.valid)
            db._close()
            self.assertFalse(db._con.valid)

    def test_connection(self):
        db = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertTrue(isinstance(db, SteadyDBConnection))
        self.assertTrue(hasattr(db, '_con'))
        self.assertTrue(hasattr(db, '_usage'))
        self.assertEqual(db._usage, 0)
        self.assertTrue(hasattr(db._con, 'valid'))
        self.assertTrue(db._con.valid)
        self.assertTrue(hasattr(db._con, 'cursor'))
        self.assertTrue(hasattr(db._con, 'close'))
        self.assertTrue(hasattr(db._con, 'open_cursors'))
        self.assertTrue(hasattr(db._con, 'num_uses'))
        self.assertTrue(hasattr(db._con, 'num_queries'))
        self.assertTrue(hasattr(db._con, 'session'))
        self.assertTrue(hasattr(db._con, 'database'))
        self.assertEqual(db._con.database, 'SteadyDBTestDB')
        self.assertTrue(hasattr(db._con, 'user'))
        self.assertEqual(db._con.user, 'SteadyDBTestUser')
        self.assertTrue(hasattr(db, 'cursor'))
        self.assertTrue(hasattr(db, 'close'))
        self.assertEqual(db._con.open_cursors, 0)
        for i in range(3):
            cursor = db.cursor()
            self.assertEqual(db._con.open_cursors, 1)
            cursor.close()
            self.assertEqual(db._con.open_cursors, 0)
        cursor = []
        for i in range(3):
            cursor.append(db.cursor())
            self.assertEqual(db._con.open_cursors, i + 1)
        del cursor
        self.assertEqual(db._con.open_cursors, 0)
        cursor = db.cursor()
        self.assertTrue(hasattr(cursor, 'execute'))
        self.assertTrue(hasattr(cursor, 'fetchone'))
        self.assertTrue(hasattr(cursor, 'callproc'))
        self.assertTrue(hasattr(cursor, 'close'))
        self.assertTrue(hasattr(cursor, 'valid'))
        self.assertTrue(cursor.valid)
        self.assertEqual(db._con.open_cursors, 1)
        for i in range(3):
            self.assertEqual(db._usage, i)
            self.assertEqual(db._con.num_uses, i)
            self.assertEqual(db._con.num_queries, i)
            cursor.execute(f'select test{i}')
            self.assertEqual(cursor.fetchone(), f'test{i}')
        self.assertTrue(cursor.valid)
        self.assertEqual(db._con.open_cursors, 1)
        for i in range(4):
            cursor.callproc('test')
        cursor.close()
        self.assertFalse(cursor.valid)
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 7)
        self.assertEqual(db._con.num_uses, 7)
        self.assertEqual(db._con.num_queries, 3)
        cursor.close()
        cursor.execute('select test8')
        self.assertTrue(cursor.valid)
        self.assertEqual(db._con.open_cursors, 1)
        self.assertEqual(cursor.fetchone(), 'test8')
        self.assertEqual(db._usage, 8)
        self.assertEqual(db._con.num_uses, 8)
        self.assertEqual(db._con.num_queries, 4)
        self.assertTrue(db._con.valid)
        db.close()
        self.assertFalse(db._con.valid)
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 8)
        self.assertEqual(db._con.num_uses, 0)
        self.assertEqual(db._con.num_queries, 0)
        self.assertRaises(dbapi.InternalError, db._con.close)
        db.close()
        self.assertRaises(dbapi.InternalError, db._con.cursor)
        cursor = db.cursor()
        self.assertTrue(db._con.valid)
        cursor.execute('select test11')
        self.assertEqual(cursor.fetchone(), 'test11')
        cursor.execute('select test12')
        self.assertEqual(cursor.fetchone(), 'test12')
        cursor.callproc('test')
        self.assertEqual(db._usage, 3)
        self.assertEqual(db._con.num_uses, 3)
        self.assertEqual(db._con.num_queries, 2)
        cursor2 = db.cursor()
        self.assertEqual(db._con.open_cursors, 2)
        cursor2.execute('select test13')
        self.assertEqual(cursor2.fetchone(), 'test13')
        self.assertEqual(db._con.num_queries, 3)
        db.close()
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._con.num_queries, 0)
        cursor = db.cursor()
        self.assertTrue(cursor.valid)
        cursor.callproc('test')
        cursor._cursor.valid = False
        self.assertFalse(cursor.valid)
        self.assertRaises(dbapi.InternalError, cursor._cursor.callproc, 'test')
        cursor.callproc('test')
        self.assertTrue(cursor.valid)
        cursor._cursor.callproc('test')
        self.assertEqual(db._usage, 2)
        self.assertEqual(db._con.num_uses, 3)
        db._con.valid = cursor._cursor.valid = False
        cursor.callproc('test')
        self.assertTrue(cursor.valid)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db._con.num_uses, 1)
        cursor.execute('set this')
        db.commit()
        cursor.execute('set that')
        db.rollback()
        self.assertEqual(
            db._con.session, ['this', 'commit', 'that', 'rollback'])

    def test_connection_context_handler(self):
        db = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertEqual(db._con.session, [])
        with db as con:
            con.cursor().execute('select test')
        self.assertEqual(db._con.session, ['commit'])
        try:
            with db as con:
                con.cursor().execute('error')
        except dbapi.ProgrammingError:
            error = True
        else:
            error = False
        self.assertTrue(error)
        self.assertEqual(db._con.session, ['commit', 'rollback'])

    def test_cursor_context_handler(self):
        db = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertEqual(db._con.open_cursors, 0)
        with db.cursor() as cursor:
            self.assertEqual(db._con.open_cursors, 1)
            cursor.execute('select test')
            self.assertEqual(cursor.fetchone(), 'test')
        self.assertEqual(db._con.open_cursors, 0)

    def test_cursor_as_iterator_provided(self):
        db = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertEqual(db._con.open_cursors, 0)
        cursor = db.cursor()
        self.assertEqual(db._con.open_cursors, 1)
        cursor.execute('select test')
        _cursor = cursor._cursor
        try:
            assert not hasattr(_cursor, 'iter')
            _cursor.__iter__ = lambda: ['test-iter']
            assert list(iter(cursor)) == ['test']
        finally:
            del _cursor.__iter__
        cursor.close()
        self.assertEqual(db._con.open_cursors, 0)

    def test_cursor_as_iterator_created(self):
        db = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertEqual(db._con.open_cursors, 0)
        cursor = db.cursor()
        self.assertEqual(db._con.open_cursors, 1)
        cursor.execute('select test')
        assert list(iter(cursor)) == ['test']
        cursor.close()
        self.assertEqual(db._con.open_cursors, 0)

    def test_connection_creator_function(self):
        db1 = SteadyDBconnect(
            dbapi, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        db2 = SteadyDBconnect(
            dbapi.connect, 0, None, None, None, True,
            'SteadyDBTestDB', user='SteadyDBTestUser')
        self.assertEqual(db1.dbapi(), db2.dbapi())
        self.assertEqual(db1.threadsafety(), db2.threadsafety())
        self.assertEqual(db1._creator, db2._creator)
        self.assertEqual(db1._args, db2._args)
        self.assertEqual(db1._kwargs, db2._kwargs)
        db2.close()
        db1.close()

    def test_connection_maxusage(self):
        db = SteadyDBconnect(dbapi, 10)
        cursor = db.cursor()
        for i in range(100):
            cursor.execute(f'select test{i}')
            r = cursor.fetchone()
            self.assertEqual(r, f'test{i}')
            self.assertTrue(db._con.valid)
            j = i % 10 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db._con.num_uses, j)
            self.assertEqual(db._con.num_queries, j)
        self.assertEqual(db._con.open_cursors, 1)
        db.begin()
        for i in range(100):
            cursor.callproc('test')
            self.assertTrue(db._con.valid)
            if i == 49:
                db.commit()
            j = i % 10 + 1 if i > 49 else i + 11
            self.assertEqual(db._usage, j)
            self.assertEqual(db._con.num_uses, j)
            j = 0 if i > 49 else 10
            self.assertEqual(db._con.num_queries, j)
        for i in range(10):
            if i == 7:
                db._con.valid = cursor._cursor.valid = False
            cursor.execute(f'select test{i}')
            r = cursor.fetchone()
            self.assertEqual(r, f'test{i}')
            j = i % 7 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db._con.num_uses, j)
            self.assertEqual(db._con.num_queries, j)
        for i in range(10):
            if i == 5:
                db._con.valid = cursor._cursor.valid = False
            cursor.callproc('test')
            j = (i + (3 if i < 5 else -5)) % 10 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db._con.num_uses, j)
            j = 3 if i < 5 else 0
            self.assertEqual(db._con.num_queries, j)
        db.close()
        cursor.execute('select test1')
        self.assertEqual(cursor.fetchone(), 'test1')
        self.assertEqual(db._usage, 1)
        self.assertEqual(db._con.num_uses, 1)
        self.assertEqual(db._con.num_queries, 1)

    def test_connection_setsession(self):
        db = SteadyDBconnect(dbapi, 3, ('set time zone', 'set datestyle'))
        self.assertTrue(hasattr(db, '_usage'))
        self.assertEqual(db._usage, 0)
        self.assertTrue(hasattr(db._con, 'open_cursors'))
        self.assertEqual(db._con.open_cursors, 0)
        self.assertTrue(hasattr(db._con, 'num_uses'))
        self.assertEqual(db._con.num_uses, 2)
        self.assertTrue(hasattr(db._con, 'num_queries'))
        self.assertEqual(db._con.num_queries, 0)
        self.assertTrue(hasattr(db._con, 'session'))
        self.assertEqual(tuple(db._con.session), ('time zone', 'datestyle'))
        for i in range(11):
            db.cursor().execute('select test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 2)
        self.assertEqual(db._con.num_uses, 4)
        self.assertEqual(db._con.num_queries, 2)
        self.assertEqual(db._con.session, ['time zone', 'datestyle'])
        db.cursor().execute('set test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 3)
        self.assertEqual(db._con.num_uses, 5)
        self.assertEqual(db._con.num_queries, 2)
        self.assertEqual(db._con.session, ['time zone', 'datestyle', 'test'])
        db.cursor().execute('select test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db._con.num_uses, 3)
        self.assertEqual(db._con.num_queries, 1)
        self.assertEqual(db._con.session, ['time zone', 'datestyle'])
        db.cursor().execute('set test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 2)
        self.assertEqual(db._con.num_uses, 4)
        self.assertEqual(db._con.num_queries, 1)
        self.assertEqual(db._con.session, ['time zone', 'datestyle', 'test'])
        db.cursor().execute('select test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 3)
        self.assertEqual(db._con.num_uses, 5)
        self.assertEqual(db._con.num_queries, 2)
        self.assertEqual(db._con.session, ['time zone', 'datestyle', 'test'])
        db.close()
        db.cursor().execute('set test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db._con.num_uses, 3)
        self.assertEqual(db._con.num_queries, 0)
        self.assertEqual(db._con.session, ['time zone', 'datestyle', 'test'])
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(db._con.open_cursors, 0)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db._con.num_uses, 3)
        self.assertEqual(db._con.num_queries, 1)
        self.assertEqual(db._con.session, ['time zone', 'datestyle'])

    def test_connection_failures(self):
        db = SteadyDBconnect(dbapi)
        db.close()
        db.cursor()
        db = SteadyDBconnect(dbapi, failures=dbapi.InternalError)
        db.close()
        db.cursor()
        db = SteadyDBconnect(dbapi, failures=dbapi.OperationalError)
        db.close()
        self.assertRaises(dbapi.InternalError, db.cursor)
        db = SteadyDBconnect(dbapi, failures=(
            dbapi.OperationalError, dbapi.InterfaceError))
        db.close()
        self.assertRaises(dbapi.InternalError, db.cursor)
        db = SteadyDBconnect(dbapi, failures=(
            dbapi.OperationalError, dbapi.InterfaceError, dbapi.InternalError))
        db.close()
        db.cursor()

    def test_connection_failure_error(self):
        db = SteadyDBconnect(dbapi)
        cursor = db.cursor()
        db.close()
        cursor.execute('select test')
        cursor = db.cursor()
        db.close()
        self.assertRaises(dbapi.ProgrammingError, cursor.execute, 'error')

    def test_connection_set_sizes(self):
        db = SteadyDBconnect(dbapi)
        cursor = db.cursor()
        cursor.execute('get sizes')
        result = cursor.fetchone()
        self.assertEqual(result, ([], {}))
        cursor.setinputsizes([7, 42, 6])
        cursor.setoutputsize(9)
        cursor.setoutputsize(15, 3)
        cursor.setoutputsize(42, 7)
        cursor.execute('get sizes')
        result = cursor.fetchone()
        self.assertEqual(result, ([7, 42, 6], {None: 9, 3: 15, 7: 42}))
        cursor.execute('get sizes')
        result = cursor.fetchone()
        self.assertEqual(result, ([], {}))
        cursor.setinputsizes([6, 42, 7])
        cursor.setoutputsize(7)
        cursor.setoutputsize(15, 3)
        cursor.setoutputsize(42, 9)
        db.close()
        cursor.execute('get sizes')
        result = cursor.fetchone()
        self.assertEqual(result, ([6, 42, 7], {None: 7, 3: 15, 9: 42}))

    def test_connection_ping_check(self):
        Connection = dbapi.Connection
        Connection.has_ping = False
        Connection.num_pings = 0
        db = SteadyDBconnect(dbapi)
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 0)
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 0)
        self.assertIsNone(db._ping_check())
        self.assertEqual(Connection.num_pings, 1)
        db = SteadyDBconnect(dbapi, ping=7)
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 2)
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 2)
        self.assertIsNone(db._ping_check())
        self.assertEqual(Connection.num_pings, 2)
        Connection.has_ping = True
        db = SteadyDBconnect(dbapi)
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 2)
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 2)
        self.assertTrue(db._ping_check())
        self.assertEqual(Connection.num_pings, 3)
        db = SteadyDBconnect(dbapi, ping=1)
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 3)
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 3)
        self.assertTrue(db._ping_check())
        self.assertEqual(Connection.num_pings, 4)
        db.close()
        self.assertTrue(db._ping_check())
        self.assertEqual(Connection.num_pings, 5)
        db = SteadyDBconnect(dbapi, ping=7)
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 7)
        db.close()
        db.cursor().execute('select test')
        self.assertEqual(Connection.num_pings, 9)
        db = SteadyDBconnect(dbapi, ping=3)
        self.assertEqual(Connection.num_pings, 9)
        db.cursor()
        self.assertEqual(Connection.num_pings, 10)
        db.close()
        cursor = db.cursor()
        self.assertEqual(Connection.num_pings, 11)
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 11)
        db = SteadyDBconnect(dbapi, ping=5)
        self.assertEqual(Connection.num_pings, 11)
        db.cursor()
        self.assertEqual(Connection.num_pings, 11)
        db.close()
        cursor = db.cursor()
        self.assertEqual(Connection.num_pings, 11)
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 12)
        db.close()
        cursor = db.cursor()
        self.assertEqual(Connection.num_pings, 12)
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 13)
        db = SteadyDBconnect(dbapi, ping=7)
        self.assertEqual(Connection.num_pings, 13)
        db.cursor()
        self.assertEqual(Connection.num_pings, 14)
        db.close()
        cursor = db.cursor()
        self.assertEqual(Connection.num_pings, 15)
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 16)
        db.close()
        cursor = db.cursor()
        self.assertEqual(Connection.num_pings, 17)
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 18)
        db.close()
        cursor.execute('select test')
        self.assertEqual(Connection.num_pings, 20)
        Connection.has_ping = False
        Connection.num_pings = 0

    def test_begin_transaction(self):
        db = SteadyDBconnect(dbapi, database='ok')
        cursor = db.cursor()
        cursor.close()
        cursor.execute('select test12')
        self.assertEqual(cursor.fetchone(), 'test12')
        db.begin()
        cursor = db.cursor()
        cursor.close()
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test12')
        cursor.execute('select test12')
        self.assertEqual(cursor.fetchone(), 'test12')
        db.close()
        db.begin()
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test12')
        cursor.execute('select test12')
        self.assertEqual(cursor.fetchone(), 'test12')
        db.begin()
        self.assertRaises(dbapi.ProgrammingError, cursor.execute, 'error')
        cursor.close()
        cursor.execute('select test12')
        self.assertEqual(cursor.fetchone(), 'test12')

    def test_with_begin_extension(self):
        db = SteadyDBconnect(dbapi, database='ok')
        db._con._begin_called_with = None

        def begin(a, b=None, c=7):
            db._con._begin_called_with = (a, b, c)

        db._con.begin = begin
        db.begin(42, 6)
        cursor = db.cursor()
        cursor.execute('select test13')
        self.assertEqual(cursor.fetchone(), 'test13')
        self.assertEqual(db._con._begin_called_with, (42, 6, 7))

    def test_cancel_transaction(self):
        db = SteadyDBconnect(dbapi, database='ok')
        cursor = db.cursor()
        db.begin()
        cursor.execute('select test14')
        self.assertEqual(cursor.fetchone(), 'test14')
        db.cancel()
        cursor.execute('select test14')
        self.assertEqual(cursor.fetchone(), 'test14')

    def test_with_cancel_extension(self):
        db = SteadyDBconnect(dbapi, database='ok')
        db._con._cancel_called = None

        def cancel():
            db._con._cancel_called = 'yes'

        db._con.cancel = cancel
        db.begin()
        cursor = db.cursor()
        cursor.execute('select test15')
        self.assertEqual(cursor.fetchone(), 'test15')
        db.cancel()
        self.assertEqual(db._con._cancel_called, 'yes')

    def test_reset_transaction(self):
        db = SteadyDBconnect(dbapi, database='ok')
        db.begin()
        self.assertFalse(db._con.session)
        db.close()
        self.assertFalse(db._con.session)
        db = SteadyDBconnect(dbapi, database='ok', closeable=False)
        db.begin()
        self.assertFalse(db._con.session)
        db.close()
        self.assertEqual(db._con.session, ['rollback'])

    def test_commit_error(self):
        db = SteadyDBconnect(dbapi, database='ok')
        db.begin()
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        db.commit()
        self.assertEqual(db._con.session, ['commit'])
        self.assertTrue(db._con.valid)
        db.begin()
        db._con.valid = False
        con = db._con
        self.assertRaises(dbapi.InternalError, db.commit)
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        self.assertIsNot(con, db._con)
        db.begin()
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        db.commit()
        self.assertEqual(db._con.session, ['commit'])
        self.assertTrue(db._con.valid)

    def test_rollback_error(self):
        db = SteadyDBconnect(dbapi, database='ok')
        db.begin()
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        db.rollback()
        self.assertEqual(db._con.session, ['rollback'])
        self.assertTrue(db._con.valid)
        db.begin()
        db._con.valid = False
        con = db._con
        self.assertRaises(dbapi.InternalError, db.rollback)
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        self.assertIsNot(con, db._con)
        db.begin()
        self.assertFalse(db._con.session)
        self.assertTrue(db._con.valid)
        db.rollback()
        self.assertEqual(db._con.session, ['rollback'])
        self.assertTrue(db._con.valid)


if __name__ == '__main__':
    unittest.main()



"""Test the ThreadingLocal module."""

import unittest
from threading import Thread

from dbutils.persistent_db import local


class TestThreadingLocal(unittest.TestCase):

    def test_getattr(self):
        my_data = local()
        my_data.number = 42
        self.assertEqual(my_data.number, 42)

    def test_dict(self):
        my_data = local()
        my_data.number = 42
        self.assertEqual(my_data.__dict__, {'number': 42})
        my_data.__dict__.setdefault('widgets', [])
        self.assertEqual(my_data.widgets, [])

    def test_threadlocal(self):
        def f():
            items = sorted(my_data.__dict__.items())
            log.append(items)
            my_data.number = 11
            log.append(my_data.number)
        my_data = local()
        my_data.number = 42
        log = []
        thread = Thread(target=f)
        thread.start()
        thread.join()
        self.assertEqual(log, [[], 11])
        self.assertEqual(my_data.number, 42)

    def test_subclass(self):

        class MyLocal(local):
            number = 2
            initialized = 0

            def __init__(self, **kw):
                if self.initialized:
                    raise SystemError
                self.initialized = 1
                self.__dict__.update(kw)

            def squared(self):
                return self.number ** 2

        my_data = MyLocal(color='red')
        self.assertEqual(my_data.number, 2)
        self.assertEqual(my_data.color, 'red')
        del my_data.color
        self.assertEqual(my_data.squared(), 4)

        def f():
            items = sorted(my_data.__dict__.items())
            log.append(items)
            my_data.number = 7
            log.append(my_data.number)

        log = []
        thread = Thread(target=f)
        thread.start()
        thread.join()
        self.assertEqual(log, [[('color', 'red'), ('initialized', 1)], 7])
        self.assertEqual(my_data.number, 2)
        self.assertFalse(hasattr(my_data, 'color'))

        class MyLocal(local):
            __slots__ = 'number'

        my_data = MyLocal()
        my_data.number = 42
        my_data.color = 'red'
        thread = Thread(target=f)
        thread.start()
        thread.join()
        self.assertEqual(my_data.number, 7)


if __name__ == '__main__':
    unittest.main()



"""Test the SimplePooledPg module.

Note:
We don't test performance here, so the test does not predicate
whether SimplePooledPg actually will help in improving performance or not.


Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_pg  # noqa

from dbutils import simple_pooled_pg


class TestSimplePooledPg(unittest.TestCase):

    def my_db_pool(self, maxConnections):
        return simple_pooled_pg.PooledPg(
            maxConnections, 'SimplePooledPgTestDB', 'SimplePooledPgTestUser')

    def test_version(self):
        from dbutils import __version__
        self.assertEqual(simple_pooled_pg.__version__, __version__)
        self.assertEqual(simple_pooled_pg.PooledPg.version, __version__)

    def test_create_connection(self):
        db_pool = self.my_db_pool(1)
        db = db_pool.connection()
        self.assertTrue(hasattr(db, 'query'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertEqual(db.num_queries, 0)
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'SimplePooledPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'SimplePooledPgTestUser')
        db.query('select 1')
        self.assertEqual(db.num_queries, 1)

    def test_close_connection(self):
        db_pool = self.my_db_pool(1)
        db = db_pool.connection()
        self.assertEqual(db.num_queries, 0)
        db.query('select 1')
        self.assertEqual(db.num_queries, 1)
        db.close()
        self.assertFalse(hasattr(db, 'num_queries'))
        db = db_pool.connection()
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'SimplePooledPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'SimplePooledPgTestUser')
        self.assertEqual(db.num_queries, 1)
        db.query('select 1')
        self.assertEqual(db.num_queries, 2)

    def test_two_connections(self):
        db_pool = self.my_db_pool(2)
        db1 = db_pool.connection()
        for i in range(5):
            db1.query('select 1')
        db2 = db_pool.connection()
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        for i in range(7):
            db2.query('select 1')
        self.assertEqual(db1.num_queries, 5)
        self.assertEqual(db2.num_queries, 7)
        db1.close()
        db1 = db_pool.connection()
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        self.assertTrue(hasattr(db1, 'query'))
        for i in range(3):
            db1.query('select 1')
        self.assertEqual(db1.num_queries, 8)
        db2.query('select 1')
        self.assertEqual(db2.num_queries, 8)

    def test_threads(self):
        db_pool = self.my_db_pool(2)
        from queue import Queue, Empty
        queue = Queue(3)

        def connection():
            queue.put(db_pool.connection())

        from threading import Thread
        threads = [Thread(target=connection).start() for i in range(3)]
        self.assertEqual(len(threads), 3)
        try:
            db1 = queue.get(1, 1)
            db2 = queue.get(1, 1)
        except TypeError:
            db1 = queue.get(1)
            db2 = queue.get(1)
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        try:
            self.assertRaises(Empty, queue.get, 1, 0.1)
        except TypeError:
            self.assertRaises(Empty, queue.get, 0)
        db2.close()
        try:
            db3 = queue.get(1, 1)
        except TypeError:
            db3 = queue.get(1)
        self.assertNotEqual(db1, db3)
        self.assertNotEqual(db1._con, db3._con)


if __name__ == '__main__':
    unittest.main()



"""Test the SimplePooledDB module.

Note:
We don't test performance here, so the test does not predicate
whether SimplePooledDB actually will help in improving performance or not.
We also do not test any real world DB-API 2 module, we just
mock the basic connection functionality of an arbitrary module.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_db as dbapi

from dbutils import simple_pooled_db


class TestSimplePooledDB(unittest.TestCase):

    def my_db_pool(self, mythreadsafety, maxConnections):
        threadsafety = dbapi.threadsafety
        dbapi.threadsafety = mythreadsafety
        try:
            return simple_pooled_db.PooledDB(
                dbapi, maxConnections,
                'SimplePooledDBTestDB', 'SimplePooledDBTestUser')
        finally:
            dbapi.threadsafety = threadsafety

    def test_version(self):
        from dbutils import __version__
        self.assertEqual(simple_pooled_db.__version__, __version__)
        self.assertEqual(simple_pooled_db.PooledDB.version, __version__)

    def test_no_threadsafety(self):
        for threadsafety in (None, -1, 0, 4):
            self.assertRaises(
                simple_pooled_db.NotSupportedError,
                self.my_db_pool, threadsafety, 1)

    def test_create_connection(self):
        for threadsafety in (1, 2, 3):
            dbpool = self.my_db_pool(threadsafety, 1)
            db = dbpool.connection()
            self.assertTrue(hasattr(db, 'cursor'))
            self.assertTrue(hasattr(db, 'open_cursors'))
            self.assertEqual(db.open_cursors, 0)
            self.assertTrue(hasattr(db, 'database'))
            self.assertEqual(db.database, 'SimplePooledDBTestDB')
            self.assertTrue(hasattr(db, 'user'))
            self.assertEqual(db.user, 'SimplePooledDBTestUser')
            cursor = db.cursor()
            self.assertIsNotNone(cursor)
            self.assertEqual(db.open_cursors, 1)
            del cursor

    def test_close_connection(self):
        for threadsafety in (1, 2, 3):
            db_pool = self.my_db_pool(threadsafety, 1)
            db = db_pool.connection()
            self.assertEqual(db.open_cursors, 0)
            cursor1 = db.cursor()
            self.assertIsNotNone(cursor1)
            self.assertEqual(db.open_cursors, 1)
            db.close()
            self.assertFalse(hasattr(db, 'open_cursors'))
            db = db_pool.connection()
            self.assertTrue(hasattr(db, 'database'))
            self.assertEqual(db.database, 'SimplePooledDBTestDB')
            self.assertTrue(hasattr(db, 'user'))
            self.assertEqual(db.user, 'SimplePooledDBTestUser')
            self.assertEqual(db.open_cursors, 1)
            cursor2 = db.cursor()
            self.assertIsNotNone(cursor2)
            self.assertEqual(db.open_cursors, 2)
            del cursor2
            del cursor1

    def test_two_connections(self):
        for threadsafety in (1, 2, 3):
            db_pool = self.my_db_pool(threadsafety, 2)
            db1 = db_pool.connection()
            cursors1 = [db1.cursor() for i in range(5)]
            db2 = db_pool.connection()
            self.assertNotEqual(db1, db2)
            cursors2 = [db2.cursor() for i in range(7)]
            self.assertEqual(db1.open_cursors, 5)
            self.assertEqual(db2.open_cursors, 7)
            db1.close()
            db1 = db_pool.connection()
            self.assertNotEqual(db1, db2)
            self.assertTrue(hasattr(db1, 'cursor'))
            for i in range(3):
                cursors1.append(db1.cursor())
            self.assertEqual(db1.open_cursors, 8)
            cursors2.append(db2.cursor())
            self.assertEqual(db2.open_cursors, 8)
            del cursors2
            del cursors1

    def test_threadsafety_1(self):
        db_pool = self.my_db_pool(1, 2)
        from queue import Queue, Empty
        queue = Queue(3)

        def connection():
            queue.put(db_pool.connection())

        from threading import Thread
        threads = [Thread(target=connection).start() for i in range(3)]
        self.assertEqual(len(threads), 3)
        try:
            db1 = queue.get(1, 1)
            db2 = queue.get(1, 1)
        except TypeError:
            db1 = queue.get(1)
            db2 = queue.get(1)
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        try:
            self.assertRaises(Empty, queue.get, 1, 0.1)
        except TypeError:
            self.assertRaises(Empty, queue.get, 0)
        db2.close()
        try:
            db3 = queue.get(1, 1)
        except TypeError:
            db3 = queue.get(1)
        self.assertNotEqual(db1, db3)
        self.assertNotEqual(db1._con, db3._con)

    def test_threadsafety_2(self):
        for threadsafety in (2, 3):
            dbpool = self.my_db_pool(threadsafety, 2)
            db1 = dbpool.connection()
            db2 = dbpool.connection()
            cursors = [dbpool.connection().cursor() for i in range(100)]
            self.assertEqual(db1.open_cursors, 50)
            self.assertEqual(db2.open_cursors, 50)
            del cursors


if __name__ == '__main__':
    unittest.main()



"""Test the PooledPg module.

Note:
We don't test performance here, so the test does not predicate
whether PooledPg actually will help in improving performance or not.
We also assume that the underlying SteadyPg connections are tested.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_pg  # noqa

from dbutils.pooled_pg import PooledPg, InvalidConnection, TooManyConnections


class TestPooledPg(unittest.TestCase):

    def test_version(self):
        from dbutils import __version__, pooled_pg
        self.assertEqual(pooled_pg.__version__, __version__)
        self.assertEqual(PooledPg.version, __version__)

    def test_create_connection(self):
        pool = PooledPg(
            1, 1, 0, False, None, None, False,
            'PooledPgTestDB', user='PooledPgTestUser')
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 1)
        self.assertTrue(hasattr(pool, '_maxusage'))
        self.assertIsNone(pool._maxusage)
        self.assertTrue(hasattr(pool, '_setsession'))
        self.assertIsNone(pool._setsession)
        self.assertTrue(hasattr(pool, '_reset'))
        self.assertFalse(pool._reset)
        db_con = pool._cache.get(0)
        pool._cache.put(db_con, 0)
        from dbutils.steady_pg import SteadyPgConnection
        self.assertTrue(isinstance(db_con, SteadyPgConnection))
        db = pool.connection()
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertTrue(hasattr(db, '_con'))
        self.assertEqual(db._con, db_con)
        self.assertTrue(hasattr(db, 'query'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertEqual(db.num_queries, 0)
        self.assertTrue(hasattr(db, '_maxusage'))
        self.assertEqual(db._maxusage, 0)
        self.assertTrue(hasattr(db, '_setsession_sql'))
        self.assertIsNone(db._setsession_sql)
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'PooledPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'PooledPgTestUser')
        db.query('select test')
        self.assertEqual(db.num_queries, 1)
        pool = PooledPg(1)
        db = pool.connection()
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertIsNone(db.dbname)
        self.assertTrue(hasattr(db, 'user'))
        self.assertIsNone(db.user)
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertEqual(db.num_queries, 0)
        pool = PooledPg(0, 0, 0, False, 3, ('set datestyle',),)
        self.assertEqual(pool._maxusage, 3)
        self.assertEqual(pool._setsession, ('set datestyle',))
        db = pool.connection()
        self.assertEqual(db._maxusage, 3)
        self.assertEqual(db._setsession_sql, ('set datestyle',))

    def test_close_connection(self):
        pool = PooledPg(
            0, 1, 0, False, None, None, False,
            'PooledPgTestDB', user='PooledPgTestUser')
        db = pool.connection()
        self.assertTrue(hasattr(db, '_con'))
        db_con = db._con
        from dbutils.steady_pg import SteadyPgConnection
        self.assertTrue(isinstance(db_con, SteadyPgConnection))
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertEqual(db.num_queries, 0)
        db.query('select test')
        self.assertEqual(db.num_queries, 1)
        db.close()
        self.assertRaises(InvalidConnection, getattr, db, 'num_queries')
        db = pool.connection()
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'PooledPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'PooledPgTestUser')
        self.assertEqual(db.num_queries, 1)
        db.query('select test')
        self.assertEqual(db.num_queries, 2)
        db = pool.connection()
        self.assertEqual(pool._cache.qsize(), 1)
        self.assertEqual(pool._cache.get(0), db_con)

    def test_min_max_cached(self):
        pool = PooledPg(3)
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 3)
        cache = [pool.connection() for i in range(3)]
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(3):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 3)
        for i in range(6):
            cache.append(pool.connection())
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(6):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 6)
        pool = PooledPg(3, 4)
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 3)
        cache = [pool.connection() for i in range(3)]
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(3):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 3)
        for i in range(6):
            cache.append(pool.connection())
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(6):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 4)
        pool = PooledPg(3, 2)
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 3)
        cache = [pool.connection() for i in range(4)]
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(4):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 3)
        pool = PooledPg(2, 5)
        self.assertTrue(hasattr(pool, '_cache'))
        self.assertEqual(pool._cache.qsize(), 2)
        cache = [pool.connection() for i in range(10)]
        self.assertEqual(pool._cache.qsize(), 0)
        for i in range(10):
            cache.pop().close()
        self.assertEqual(pool._cache.qsize(), 5)

    def test_max_connections(self):
        from dbutils.pooled_pg import TooManyConnections
        pool = PooledPg(1, 2, 3)
        self.assertEqual(pool._cache.qsize(), 1)
        cache = [pool.connection() for i in range(3)]
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertRaises(TooManyConnections, pool.connection)
        pool = PooledPg(0, 1, 1, False)
        self.assertEqual(pool._blocking, 0)
        self.assertEqual(pool._cache.qsize(), 0)
        db = pool.connection()
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertRaises(TooManyConnections, pool.connection)
        del db
        del cache
        pool = PooledPg(1, 2, 1)
        self.assertEqual(pool._cache.qsize(), 1)
        cache = [pool.connection()]
        self.assertEqual(pool._cache.qsize(), 0)
        cache.append(pool.connection())
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertRaises(TooManyConnections, pool.connection)
        pool = PooledPg(3, 2, 1, False)
        self.assertEqual(pool._cache.qsize(), 3)
        cache = [pool.connection() for i in range(3)]
        self.assertEqual(len(cache), 3)
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertRaises(TooManyConnections, pool.connection)
        pool = PooledPg(1, 1, 1, True)
        self.assertEqual(pool._blocking, 1)
        self.assertEqual(pool._cache.qsize(), 1)
        db = pool.connection()
        self.assertEqual(pool._cache.qsize(), 0)

        def connection():
            pool.connection().query('set thread')

        from threading import Thread
        thread = Thread(target=connection)
        thread.start()
        thread.join(0.1)
        self.assertTrue(thread.is_alive())
        self.assertEqual(pool._cache.qsize(), 0)
        session = db._con.session
        self.assertEqual(session, [])
        del db
        thread.join(0.1)
        self.assertFalse(thread.is_alive())
        self.assertEqual(pool._cache.qsize(), 1)
        db = pool.connection()
        self.assertEqual(pool._cache.qsize(), 0)
        self.assertEqual(session, ['thread'])
        del db

    def test_one_thread_two_connections(self):
        pool = PooledPg(2)
        db1 = pool.connection()
        for i in range(5):
            db1.query('select test')
        db2 = pool.connection()
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        for i in range(7):
            db2.query('select test')
        self.assertEqual(db1.num_queries, 5)
        self.assertEqual(db2.num_queries, 7)
        del db1
        db1 = pool.connection()
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        self.assertTrue(hasattr(db1, 'query'))
        for i in range(3):
            db1.query('select test')
        self.assertEqual(db1.num_queries, 8)
        db2.query('select test')
        self.assertEqual(db2.num_queries, 8)

    def test_three_threads_two_connections(self):
        pool = PooledPg(2, 2, 2, True)
        from queue import Queue, Empty
        queue = Queue(3)

        def connection():
            try:
                queue.put(pool.connection(), 1, 1)
            except TypeError:
                queue.put(pool.connection(), 1)

        from threading import Thread
        for i in range(3):
            Thread(target=connection).start()
        try:
            db1 = queue.get(1, 1)
            db2 = queue.get(1, 1)
        except TypeError:
            db1 = queue.get(1)
            db2 = queue.get(1)
        db1_con = db1._con
        db2_con = db2._con
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1_con, db2_con)
        try:
            self.assertRaises(Empty, queue.get, 1, 0.1)
        except TypeError:
            self.assertRaises(Empty, queue.get, 0)
        del db1
        try:
            db1 = queue.get(1, 1)
        except TypeError:
            db1 = queue.get(1)
        self.assertNotEqual(db1, db2)
        self.assertNotEqual(db1._con, db2._con)
        self.assertEqual(db1._con, db1_con)

    def test_reset_transaction(self):
        pool = PooledPg(1)
        db = pool.connection()
        db.begin()
        con = db._con
        self.assertTrue(con._transaction)
        db.query('select test')
        self.assertEqual(con.num_queries, 1)
        db.close()
        self.assertIs(pool.connection()._con, con)
        self.assertFalse(con._transaction)
        self.assertEqual(con.session, ['begin', 'rollback'])
        self.assertEqual(con.num_queries, 1)
        pool = PooledPg(1, reset=1)
        db = pool.connection()
        db.begin()
        con = db._con
        self.assertTrue(con._transaction)
        self.assertEqual(con.session, ['rollback', 'begin'])
        db.query('select test')
        self.assertEqual(con.num_queries, 1)
        db.close()
        self.assertIs(pool.connection()._con, con)
        self.assertFalse(con._transaction)
        self.assertEqual(
            con.session, ['rollback', 'begin', 'rollback', 'rollback'])
        self.assertEqual(con.num_queries, 1)
        pool = PooledPg(1, reset=2)
        db = pool.connection()
        db.begin()
        con = db._con
        self.assertTrue(con._transaction)
        self.assertEqual(con.session, ['begin'])
        db.query('select test')
        self.assertEqual(con.num_queries, 1)
        db.close()
        self.assertIs(pool.connection()._con, con)
        self.assertFalse(con._transaction)
        self.assertEqual(con.session, [])
        self.assertEqual(con.num_queries, 0)

    def test_context_manager(self):
        pool = PooledPg(1, 1, 1)
        with pool.connection() as db:
            db_con = db._con._con
            db.query('select test')
            self.assertEqual(db_con.num_queries, 1)
            self.assertRaises(TooManyConnections, pool.connection)
        with pool.connection() as db:
            db_con = db._con._con
            db.query('select test')
            self.assertEqual(db_con.num_queries, 2)
            self.assertRaises(TooManyConnections, pool.connection)


if __name__ == '__main__':
    unittest.main()



"""Test the PooledDB module.

Note:
We don't test performance here, so the test does not predicate
whether PooledDB actually will help in improving performance or not.
We also assume that the underlying SteadyDB connections are tested.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_db as dbapi

from dbutils.pooled_db import (
    PooledDB, SharedDBConnection, InvalidConnection, TooManyConnections)


class TestPooledDB(unittest.TestCase):

    def test_version(self):
        from dbutils import __version__, pooled_db
        self.assertEqual(pooled_db.__version__, __version__)
        self.assertEqual(PooledDB.version, __version__)

    def test_no_threadsafety(self):
        from dbutils.pooled_db import NotSupportedError
        for threadsafety in (None, 0):
            dbapi.threadsafety = threadsafety
            self.assertRaises(NotSupportedError, PooledDB, dbapi)

    def test_threadsafety(self):
        for threadsafety in (1, 2, 3):
            dbapi.threadsafety = threadsafety
            pool = PooledDB(dbapi, 0, 0, 1)
            self.assertTrue(hasattr(pool, '_maxshared'))
            if threadsafety > 1:
                self.assertEqual(pool._maxshared, 1)
                self.assertTrue(hasattr(pool, '_shared_cache'))
            else:
                self.assertEqual(pool._maxshared, 0)
                self.assertFalse(hasattr(pool, '_shared_cache'))

    def test_create_connection(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(
                dbapi, 1, 1, 1, 0, False, None, None, True, None, None,
                'PooledDBTestDB', user='PooledDBTestUser')
            self.assertTrue(hasattr(pool, '_idle_cache'))
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertTrue(hasattr(pool, '_shared_cache'))
                self.assertEqual(len(pool._shared_cache), 0)
            else:
                self.assertFalse(hasattr(pool, '_shared_cache'))
            self.assertTrue(hasattr(pool, '_maxusage'))
            self.assertIsNone(pool._maxusage)
            self.assertTrue(hasattr(pool, '_setsession'))
            self.assertIsNone(pool._setsession)
            con = pool._idle_cache[0]
            from dbutils.steady_db import SteadyDBConnection
            self.assertTrue(isinstance(con, SteadyDBConnection))
            self.assertTrue(hasattr(con, '_maxusage'))
            self.assertEqual(con._maxusage, 0)
            self.assertTrue(hasattr(con, '_setsession_sql'))
            self.assertIsNone(con._setsession_sql)
            db = pool.connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            self.assertTrue(hasattr(db, '_con'))
            self.assertEqual(db._con, con)
            self.assertTrue(hasattr(db, 'cursor'))
            self.assertTrue(hasattr(db, '_usage'))
            self.assertEqual(db._usage, 0)
            self.assertTrue(hasattr(con, '_con'))
            db_con = con._con
            self.assertTrue(hasattr(db_con, 'database'))
            self.assertEqual(db_con.database, 'PooledDBTestDB')
            self.assertTrue(hasattr(db_con, 'user'))
            self.assertEqual(db_con.user, 'PooledDBTestUser')
            self.assertTrue(hasattr(db_con, 'open_cursors'))
            self.assertEqual(db_con.open_cursors, 0)
            self.assertTrue(hasattr(db_con, 'num_uses'))
            self.assertEqual(db_con.num_uses, 0)
            self.assertTrue(hasattr(db_con, 'num_queries'))
            self.assertEqual(db_con.num_queries, 0)
            cursor = db.cursor()
            self.assertEqual(db_con.open_cursors, 1)
            cursor.execute('select test')
            r = cursor.fetchone()
            cursor.close()
            self.assertEqual(db_con.open_cursors, 0)
            self.assertEqual(r, 'test')
            self.assertEqual(db_con.num_queries, 1)
            self.assertEqual(db._usage, 1)
            cursor = db.cursor()
            self.assertEqual(db_con.open_cursors, 1)
            cursor.execute('set sessiontest')
            cursor2 = db.cursor()
            self.assertEqual(db_con.open_cursors, 2)
            cursor2.close()
            self.assertEqual(db_con.open_cursors, 1)
            cursor.close()
            self.assertEqual(db_con.open_cursors, 0)
            self.assertEqual(db_con.num_queries, 1)
            self.assertEqual(db._usage, 2)
            self.assertEqual(
                db_con.session, ['rollback', 'sessiontest'])
            pool = PooledDB(dbapi, 1, 1, 1)
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection(True)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection(False)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            self.assertEqual(db._usage, 0)
            db_con = db._con._con
            self.assertIsNone(db_con.database)
            self.assertIsNone(db_con.user)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.dedicated_connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            self.assertEqual(db._usage, 0)
            db_con = db._con._con
            self.assertIsNone(db_con.database)
            self.assertIsNone(db_con.user)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            pool = PooledDB(dbapi, 0, 0, 0, 0, False, 3, ('set datestyle',))
            self.assertEqual(pool._maxusage, 3)
            self.assertEqual(pool._setsession, ('set datestyle',))
            con = pool.connection()._con
            self.assertEqual(con._maxusage, 3)
            self.assertEqual(con._setsession_sql, ('set datestyle',))

    def test_close_connection(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(
                dbapi, 0, 1, 1, 0, False, None, None, True, None, None,
                'PooledDBTestDB', user='PooledDBTestUser')
            self.assertTrue(hasattr(pool, '_idle_cache'))
            self.assertEqual(len(pool._idle_cache), 0)
            db = pool.connection()
            self.assertTrue(hasattr(db, '_con'))
            con = db._con
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
                self.assertTrue(hasattr(db, '_shared_con'))
                shared_con = db._shared_con
                self.assertEqual(pool._shared_cache[0], shared_con)
                self.assertTrue(hasattr(shared_con, 'shared'))
                self.assertEqual(shared_con.shared, 1)
                self.assertTrue(hasattr(shared_con, 'con'))
                self.assertEqual(shared_con.con, con)
            from dbutils.steady_db import SteadyDBConnection
            self.assertTrue(isinstance(con, SteadyDBConnection))
            self.assertTrue(hasattr(con, '_con'))
            db_con = con._con
            self.assertTrue(hasattr(db_con, 'num_queries'))
            self.assertEqual(db._usage, 0)
            self.assertEqual(db_con.num_queries, 0)
            db.cursor().execute('select test')
            self.assertEqual(db._usage, 1)
            self.assertEqual(db_con.num_queries, 1)
            db.close()
            self.assertIsNone(db._con)
            if shareable:
                self.assertIsNone(db._shared_con)
                self.assertEqual(shared_con.shared, 0)
            self.assertRaises(InvalidConnection, getattr, db, '_usage')
            self.assertFalse(hasattr(db_con, '_num_queries'))
            self.assertEqual(len(pool._idle_cache), 1)
            self.assertEqual(pool._idle_cache[0]._con, db_con)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db.close()
            if shareable:
                self.assertEqual(shared_con.shared, 0)
            db = pool.connection()
            self.assertEqual(db._con, con)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
                shared_con = db._shared_con
                self.assertEqual(pool._shared_cache[0], shared_con)
                self.assertEqual(shared_con.con, con)
                self.assertEqual(shared_con.shared, 1)
            self.assertEqual(db._usage, 1)
            self.assertEqual(db_con.num_queries, 1)
            self.assertTrue(hasattr(db_con, 'database'))
            self.assertEqual(db_con.database, 'PooledDBTestDB')
            self.assertTrue(hasattr(db_con, 'user'))
            self.assertEqual(db_con.user, 'PooledDBTestUser')
            db.cursor().execute('select test')
            self.assertEqual(db_con.num_queries, 2)
            db.cursor().execute('select test')
            self.assertEqual(db_con.num_queries, 3)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            self.assertEqual(pool._idle_cache[0]._con, db_con)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection(False)
            self.assertEqual(db._con, con)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db.close()
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)

    def test_close_all(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 10)
            self.assertEqual(len(pool._idle_cache), 10)
            pool.close()
            self.assertEqual(len(pool._idle_cache), 0)
            pool = PooledDB(dbapi, 10)
            closed = ['no']

            def close(what=closed):
                what[0] = 'yes'

            pool._idle_cache[7]._con.close = close
            self.assertEqual(closed, ['no'])
            del pool
            self.assertEqual(closed, ['yes'])
            pool = PooledDB(dbapi, 10, 10, 5)
            self.assertEqual(len(pool._idle_cache), 10)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            cache = []
            for i in range(5):
                cache.append(pool.connection())
            self.assertEqual(len(pool._idle_cache), 5)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 5)
            else:
                self.assertEqual(len(pool._idle_cache), 5)
            pool.close()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            pool = PooledDB(dbapi, 10, 10, 5)
            closed = []

            def close_idle(what=closed):
                what.append('idle')

            def close_shared(what=closed):
                what.append('shared')

            if shareable:
                cache = []
                for i in range(5):
                    cache.append(pool.connection())
                pool._shared_cache[3].con.close = close_shared
            else:
                pool._idle_cache[7]._con.close = close_shared
            pool._idle_cache[3]._con.close = close_idle
            self.assertEqual(closed, [])
            del pool
            if shareable:
                del cache
            self.assertEqual(closed, ['idle', 'shared'])

    def test_shareable_connection(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 0, 1, 2)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db1 = pool.connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            db2 = pool.connection()
            self.assertNotEqual(db1._con, db2._con)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 2)
            db3 = pool.connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 2)
                self.assertEqual(db3._con, db1._con)
                self.assertEqual(db1._shared_con.shared, 2)
                self.assertEqual(db2._shared_con.shared, 1)
            else:
                self.assertNotEqual(db3._con, db1._con)
                self.assertNotEqual(db3._con, db2._con)
            db4 = pool.connection()
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 2)
                self.assertEqual(db4._con, db2._con)
                self.assertEqual(db1._shared_con.shared, 2)
                self.assertEqual(db2._shared_con.shared, 2)
            else:
                self.assertNotEqual(db4._con, db1._con)
                self.assertNotEqual(db4._con, db2._con)
                self.assertNotEqual(db4._con, db3._con)
            db5 = pool.connection(False)
            self.assertNotEqual(db5._con, db1._con)
            self.assertNotEqual(db5._con, db2._con)
            self.assertNotEqual(db5._con, db3._con)
            self.assertNotEqual(db5._con, db4._con)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 2)
                self.assertEqual(db1._shared_con.shared, 2)
                self.assertEqual(db2._shared_con.shared, 2)
            db5.close()
            self.assertEqual(len(pool._idle_cache), 1)
            db5 = pool.connection()
            if shareable:
                self.assertEqual(len(pool._idle_cache), 1)
                self.assertEqual(len(pool._shared_cache), 2)
                self.assertEqual(db5._shared_con.shared, 3)
            else:
                self.assertEqual(len(pool._idle_cache), 0)
            pool = PooledDB(dbapi, 0, 0, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            db1 = pool.connection(False)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db2 = pool.connection()
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            db3 = pool.connection()
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
                self.assertEqual(db2._con, db3._con)
            else:
                self.assertNotEqual(db2._con, db3._con)
            del db3
            if shareable:
                self.assertEqual(len(pool._idle_cache), 0)
                self.assertEqual(len(pool._shared_cache), 1)
            else:
                self.assertEqual(len(pool._idle_cache), 1)
            del db2
            if shareable:
                self.assertEqual(len(pool._idle_cache), 1)
                self.assertEqual(len(pool._shared_cache), 0)
            else:
                self.assertEqual(len(pool._idle_cache), 2)
            del db1
            if shareable:
                self.assertEqual(len(pool._idle_cache), 2)
                self.assertEqual(len(pool._shared_cache), 0)
            else:
                self.assertEqual(len(pool._idle_cache), 3)

    def test_min_max_cached(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 3)
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(3)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(6)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 6)
            pool = PooledDB(dbapi, 0, 3)
            self.assertEqual(len(pool._idle_cache), 0)
            cache = [pool.connection() for i in range(3)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(6)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            pool = PooledDB(dbapi, 3, 3)
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(3)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(6)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            pool = PooledDB(dbapi, 3, 2)
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(4)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            pool = PooledDB(dbapi, 2, 5)
            self.assertEqual(len(pool._idle_cache), 2)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 5)
            pool = PooledDB(dbapi, 1, 2, 3)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = [pool.connection(False) for i in range(4)]
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 2)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 3)
            del cache
            self.assertEqual(len(pool._idle_cache), 2)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            pool = PooledDB(dbapi, 1, 3, 2)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = [pool.connection(False) for i in range(4)]
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            cache = [pool.connection() for i in range(10)]
            if shareable:
                self.assertEqual(len(pool._idle_cache), 1)
                self.assertEqual(len(pool._shared_cache), 2)
            else:
                self.assertEqual(len(pool._idle_cache), 0)
            del cache
            self.assertEqual(len(pool._idle_cache), 3)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)

    def test_max_shared(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi)
            self.assertEqual(len(pool._idle_cache), 0)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(cache), 10)
            self.assertEqual(len(pool._idle_cache), 0)
            pool = PooledDB(dbapi, 1, 1, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(cache), 10)
            self.assertEqual(len(pool._idle_cache), 0)
            pool = PooledDB(dbapi, 0, 0, 1)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(cache), 10)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            pool = PooledDB(dbapi, 1, 1, 1)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(cache), 10)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
            pool = PooledDB(dbapi, 0, 0, 7)
            cache = [pool.connection(False) for i in range(3)]
            self.assertEqual(len(cache), 3)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            cache = [pool.connection() for i in range(10)]
            self.assertEqual(len(cache), 10)
            self.assertEqual(len(pool._idle_cache), 3)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 7)

    def test_sort_shared(self):
        dbapi.threadsafety = 2
        pool = PooledDB(dbapi, 0, 4, 4)
        cache = []
        for i in range(6):
            db = pool.connection()
            db.cursor().execute('select test')
            cache.append(db)
        for i, db in enumerate(cache):
            self.assertEqual(db._shared_con.shared, 1 if 2 <= i < 4 else 2)
        cache[2].begin()
        cache[3].begin()
        db = pool.connection()
        self.assertIs(db._con, cache[0]._con)
        db.close()
        cache[3].rollback()
        db = pool.connection()
        self.assertIs(db._con, cache[3]._con)

    def test_equally_shared(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 5, 5, 5)
            self.assertEqual(len(pool._idle_cache), 5)
            for i in range(15):
                db = pool.connection(False)
                db.cursor().execute('select test')
                db.close()
            self.assertEqual(len(pool._idle_cache), 5)
            for i in range(5):
                con = pool._idle_cache[i]
                self.assertEqual(con._usage, 3)
                self.assertEqual(con._con.num_queries, 3)
            cache = []
            for i in range(35):
                db = pool.connection()
                db.cursor().execute('select test')
                cache.append(db)
                del db
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 5)
                for i in range(5):
                    con = pool._shared_cache[i]
                    self.assertEqual(con.shared, 7)
                    con = con.con
                    self.assertEqual(con._usage, 10)
                    self.assertEqual(con._con.num_queries, 10)
            del cache
            self.assertEqual(len(pool._idle_cache), 5)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)

    def test_many_shared(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 0, 0, 5)
            cache = []
            for i in range(35):
                db = pool.connection()
                db.cursor().execute('select test1')
                db.cursor().execute('select test2')
                db.cursor().callproc('test3')
                cache.append(db)
                del db
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 5)
                for i in range(5):
                    con = pool._shared_cache[i]
                    self.assertEqual(con.shared, 7)
                    con = con.con
                    self.assertEqual(con._usage, 21)
                    self.assertEqual(con._con.num_queries, 14)
                cache[3] = cache[8] = cache[33] = None
                cache[12] = cache[17] = cache[34] = None
                self.assertEqual(len(pool._shared_cache), 5)
                self.assertEqual(pool._shared_cache[0].shared, 7)
                self.assertEqual(pool._shared_cache[1].shared, 7)
                self.assertEqual(pool._shared_cache[2].shared, 5)
                self.assertEqual(pool._shared_cache[3].shared, 4)
                self.assertEqual(pool._shared_cache[4].shared, 6)
                for db in cache:
                    if db:
                        db.cursor().callproc('test4')
                for i in range(6):
                    db = pool.connection()
                    db.cursor().callproc('test4')
                    cache.append(db)
                    del db
                for i in range(5):
                    con = pool._shared_cache[i]
                    self.assertEqual(con.shared, 7)
                    con = con.con
                    self.assertEqual(con._usage, 28)
                    self.assertEqual(con._con.num_queries, 14)
            del cache
            if shareable:
                self.assertEqual(len(pool._idle_cache), 5)
                self.assertEqual(len(pool._shared_cache), 0)
            else:
                self.assertEqual(len(pool._idle_cache), 35)

    def test_rollback(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            pool = PooledDB(dbapi, 0, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            db = pool.connection(False)
            self.assertEqual(len(pool._idle_cache), 0)
            self.assertEqual(db._con._con.open_cursors, 0)
            cursor = db.cursor()
            self.assertEqual(db._con._con.open_cursors, 1)
            cursor.execute('set doit1')
            db.commit()
            cursor.execute('set dont1')
            cursor.close()
            self.assertEqual(db._con._con.open_cursors, 0)
            del db
            self.assertEqual(len(pool._idle_cache), 1)
            db = pool.connection(False)
            self.assertEqual(len(pool._idle_cache), 0)
            self.assertEqual(db._con._con.open_cursors, 0)
            cursor = db.cursor()
            self.assertEqual(db._con._con.open_cursors, 1)
            cursor.execute('set doit2')
            cursor.close()
            self.assertEqual(db._con._con.open_cursors, 0)
            db.commit()
            session = db._con._con.session
            db.close()
            self.assertEqual(session, [
                'doit1', 'commit', 'dont1', 'rollback',
                'doit2', 'commit', 'rollback'])

    def test_maxconnections(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 1, 2, 2, 3)
            self.assertTrue(hasattr(pool, '_maxconnections'))
            self.assertEqual(pool._maxconnections, 3)
            self.assertTrue(hasattr(pool, '_connections'))
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = []
            for i in range(3):
                cache.append(pool.connection(False))
            self.assertEqual(pool._connections, 3)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            self.assertRaises(TooManyConnections, pool.connection)
            cache = []
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 2)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            for i in range(3):
                cache.append(pool.connection())
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(pool._connections, 2)
                self.assertEqual(len(pool._shared_cache), 2)
                cache.append(pool.connection(False))
                self.assertEqual(pool._connections, 3)
                self.assertEqual(len(pool._shared_cache), 2)
            else:
                self.assertEqual(pool._connections, 3)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            if shareable:
                cache.append(pool.connection(True))
                self.assertEqual(pool._connections, 3)
            else:
                self.assertRaises(TooManyConnections, pool.connection)
            del cache
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 2)
            pool = PooledDB(dbapi, 0, 1, 1, 1)
            self.assertEqual(pool._maxconnections, 1)
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 0)
            db = pool.connection(False)
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            self.assertRaises(TooManyConnections, pool.connection)
            del db
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = [pool.connection()]
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 1)
                cache.append(pool.connection())
                self.assertEqual(pool._connections, 1)
                self.assertEqual(len(pool._shared_cache), 1)
                self.assertEqual(pool._shared_cache[0].shared, 2)
            else:
                self.assertRaises(TooManyConnections, pool.connection)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            if shareable:
                cache.append(pool.connection(True))
                self.assertEqual(pool._connections, 1)
                self.assertEqual(len(pool._shared_cache), 1)
                self.assertEqual(pool._shared_cache[0].shared, 3)
            else:
                self.assertRaises(TooManyConnections, pool.connection, 1)
            del cache
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection(False)
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            del db
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            pool = PooledDB(dbapi, 1, 2, 2, 1)
            self.assertEqual(pool._maxconnections, 2)
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            cache = []
            cache.append(pool.connection(False))
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            cache.append(pool.connection(False))
            self.assertEqual(pool._connections, 2)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            self.assertRaises(TooManyConnections, pool.connection)
            pool = PooledDB(dbapi, 4, 3, 2, 1, False)
            self.assertEqual(pool._maxconnections, 4)
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 4)
            cache = []
            for i in range(4):
                cache.append(pool.connection(False))
            self.assertEqual(pool._connections, 4)
            self.assertEqual(len(pool._idle_cache), 0)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            self.assertRaises(TooManyConnections, pool.connection)
            pool = PooledDB(dbapi, 1, 2, 3, 4, False)
            self.assertEqual(pool._maxconnections, 4)
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            for i in range(4):
                cache.append(pool.connection())
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(pool._connections, 3)
                self.assertEqual(len(pool._shared_cache), 3)
                cache.append(pool.connection())
                self.assertEqual(pool._connections, 3)
                cache.append(pool.connection(False))
                self.assertEqual(pool._connections, 4)
            else:
                self.assertEqual(pool._connections, 4)
                self.assertRaises(TooManyConnections, pool.connection)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            pool = PooledDB(dbapi, 0, 0, 3, 3, False)
            self.assertEqual(pool._maxconnections, 3)
            self.assertEqual(pool._connections, 0)
            cache = []
            for i in range(3):
                cache.append(pool.connection(False))
            self.assertEqual(pool._connections, 3)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            self.assertRaises(TooManyConnections, pool.connection, 1)
            cache = []
            self.assertEqual(pool._connections, 0)
            for i in range(3):
                cache.append(pool.connection())
            self.assertEqual(pool._connections, 3)
            if shareable:
                for i in range(3):
                    cache.append(pool.connection())
                self.assertEqual(pool._connections, 3)
            else:
                self.assertRaises(TooManyConnections, pool.connection)
            self.assertRaises(TooManyConnections, pool.connection, 0)
            pool = PooledDB(dbapi, 0, 0, 3)
            self.assertEqual(pool._maxconnections, 0)
            self.assertEqual(pool._connections, 0)
            cache = []
            for i in range(10):
                cache.append(pool.connection(False))
                cache.append(pool.connection())
            if shareable:
                self.assertEqual(pool._connections, 13)
                self.assertEqual(len(pool._shared_cache), 3)
            else:
                self.assertEqual(pool._connections, 20)
            pool = PooledDB(dbapi, 1, 1, 1, 1, True)
            self.assertEqual(pool._maxconnections, 1)
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            db = pool.connection(False)
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)

            def connection():
                db = pool.connection()
                cursor = db.cursor()
                cursor.execute('set thread')
                cursor.close()
                db.close()

            from threading import Thread
            thread = Thread(target=connection)
            thread.start()
            thread.join(0.1)
            self.assertTrue(thread.is_alive())
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            session = db._con._con.session
            self.assertEqual(session, ['rollback'])
            del db
            thread.join(0.1)
            self.assertFalse(thread.is_alive())
            self.assertEqual(pool._connections, 0)
            self.assertEqual(len(pool._idle_cache), 1)
            if shareable:
                self.assertEqual(len(pool._shared_cache), 0)
            db = pool.connection(False)
            self.assertEqual(pool._connections, 1)
            self.assertEqual(len(pool._idle_cache), 0)
            self.assertEqual(
                session, ['rollback', 'rollback', 'thread', 'rollback'])
            del db

    def test_maxusage(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            for maxusage in (0, 3, 7):
                pool = PooledDB(dbapi, 0, 0, 0, 1, False, maxusage)
                self.assertEqual(pool._maxusage, maxusage)
                self.assertEqual(len(pool._idle_cache), 0)
                db = pool.connection(False)
                self.assertEqual(db._con._maxusage, maxusage)
                self.assertEqual(len(pool._idle_cache), 0)
                self.assertEqual(db._con._con.open_cursors, 0)
                self.assertEqual(db._usage, 0)
                self.assertEqual(db._con._con.num_uses, 0)
                self.assertEqual(db._con._con.num_queries, 0)
                for i in range(20):
                    cursor = db.cursor()
                    self.assertEqual(db._con._con.open_cursors, 1)
                    cursor.execute(f'select test{i}')
                    r = cursor.fetchone()
                    self.assertEqual(r, f'test{i}')
                    cursor.close()
                    self.assertEqual(db._con._con.open_cursors, 0)
                    if maxusage:
                        j = i % maxusage + 1
                    else:
                        j = i + 1
                    self.assertEqual(db._usage, j)
                    self.assertEqual(db._con._con.num_uses, j)
                    self.assertEqual(db._con._con.num_queries, j)
                db.cursor().callproc('test')
                self.assertEqual(db._con._con.open_cursors, 0)
                self.assertEqual(db._usage, j + 1)
                self.assertEqual(db._con._con.num_uses, j + 1)
                self.assertEqual(db._con._con.num_queries, j)

    def test_setsession(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            setsession = ('set time zone', 'set datestyle')
            pool = PooledDB(dbapi, 0, 0, 0, 1, False, None, setsession)
            self.assertEqual(pool._setsession, setsession)
            db = pool.connection(False)
            self.assertEqual(db._setsession_sql, setsession)
            self.assertEqual(
                db._con._con.session, ['time zone', 'datestyle'])
            db.cursor().execute('select test')
            db.cursor().execute('set test1')
            self.assertEqual(db._usage, 2)
            self.assertEqual(db._con._con.num_uses, 4)
            self.assertEqual(db._con._con.num_queries, 1)
            self.assertEqual(
                db._con._con.session, ['time zone', 'datestyle', 'test1'])
            db.close()
            db = pool.connection(False)
            self.assertEqual(db._setsession_sql, setsession)
            self.assertEqual(
                db._con._con.session,
                ['time zone', 'datestyle', 'test1', 'rollback'])
            db._con._con.close()
            db.cursor().execute('select test')
            db.cursor().execute('set test2')
            self.assertEqual(
                db._con._con.session, ['time zone', 'datestyle', 'test2'])

    def test_one_thread_two_connections(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            shareable = threadsafety > 1
            pool = PooledDB(dbapi, 2)
            db1 = pool.connection()
            for i in range(5):
                db1.cursor().execute('select test')
            db2 = pool.connection()
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            for i in range(7):
                db2.cursor().execute('select test')
            self.assertEqual(db1._con._con.num_queries, 5)
            self.assertEqual(db2._con._con.num_queries, 7)
            del db1
            db1 = pool.connection()
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            for i in range(3):
                db1.cursor().execute('select test')
            self.assertEqual(db1._con._con.num_queries, 8)
            db2.cursor().execute('select test')
            self.assertEqual(db2._con._con.num_queries, 8)
            pool = PooledDB(dbapi, 0, 0, 2)
            db1 = pool.connection()
            for i in range(5):
                db1.cursor().execute('select test')
            db2 = pool.connection()
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            for i in range(7):
                db2.cursor().execute('select test')
            self.assertEqual(db1._con._con.num_queries, 5)
            self.assertEqual(db2._con._con.num_queries, 7)
            del db1
            db1 = pool.connection()
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            for i in range(3):
                db1.cursor().execute('select test')
            self.assertEqual(db1._con._con.num_queries, 8)
            db2.cursor().execute('select test')
            self.assertEqual(db2._con._con.num_queries, 8)
            pool = PooledDB(dbapi, 0, 0, 1)
            db1 = pool.connection()
            db2 = pool.connection()
            self.assertNotEqual(db1, db2)
            if shareable:
                self.assertEqual(db1._con, db2._con)
            else:
                self.assertNotEqual(db1._con, db2._con)
            del db1
            db1 = pool.connection(False)
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)

    def test_tnree_threads_two_connections(self):
        for threadsafety in (1, 2):
            dbapi.threadsafety = threadsafety
            pool = PooledDB(dbapi, 2, 2, 0, 2, True)
            from queue import Queue, Empty
            queue = Queue(3)

            def connection():
                try:
                    queue.put(pool.connection(), 1, 1)
                except Exception:
                    queue.put(pool.connection(), 1)

            from threading import Thread
            for i in range(3):
                Thread(target=connection).start()
            try:
                db1 = queue.get(1, 1)
                db2 = queue.get(1, 1)
            except TypeError:
                db1 = queue.get(1)
                db2 = queue.get(1)
            self.assertNotEqual(db1, db2)
            db1_con = db1._con
            db2_con = db2._con
            self.assertNotEqual(db1_con, db2_con)
            try:
                self.assertRaises(Empty, queue.get, 1, 0.1)
            except TypeError:
                self.assertRaises(Empty, queue.get, 0)
            del db1
            try:
                db1 = queue.get(1, 1)
            except TypeError:
                db1 = queue.get(1)
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            self.assertEqual(db1._con, db1_con)
            pool = PooledDB(dbapi, 2, 2, 1, 2, True)
            db1 = pool.connection(False)
            db2 = pool.connection(False)
            self.assertNotEqual(db1, db2)
            db1_con = db1._con
            db2_con = db2._con
            self.assertNotEqual(db1_con, db2_con)
            Thread(target=connection).start()
            try:
                self.assertRaises(Empty, queue.get, 1, 0.1)
            except TypeError:
                self.assertRaises(Empty, queue.get, 0)
            del db1
            try:
                db1 = queue.get(1, 1)
            except TypeError:
                db1 = queue.get(1)
            self.assertNotEqual(db1, db2)
            self.assertNotEqual(db1._con, db2._con)
            self.assertEqual(db1._con, db1_con)

    def test_ping_check(self):
        Connection = dbapi.Connection
        Connection.has_ping = True
        Connection.num_pings = 0
        dbapi.threadsafety = 2
        pool = PooledDB(dbapi, 1, 1, 0, 0, False, None, None, True, None, 0)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        db._con.close()
        db.close()
        db = pool.connection()
        self.assertFalse(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        pool = PooledDB(dbapi, 1, 1, 1, 0, False, None, None, True, None, 0)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        db._con.close()
        db = pool.connection()
        self.assertFalse(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        pool = PooledDB(dbapi, 1, 1, 0, 0, False, None, None, True, None, 1)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 1)
        db._con.close()
        db.close()
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 2)
        pool = PooledDB(dbapi, 1, 1, 1, 0, False, None, None, True, None, 1)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        db._con.close()
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 4)
        pool = PooledDB(dbapi, 1, 1, 1, 0, False, None, None, True, None, 2)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 4)
        db._con.close()
        db = pool.connection()
        self.assertFalse(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 4)
        db.cursor()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 5)
        pool = PooledDB(dbapi, 1, 1, 1, 0, False, None, None, True, None, 4)
        db = pool.connection()
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 5)
        db._con.close()
        db = pool.connection()
        self.assertFalse(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 5)
        cursor = db.cursor()
        db._con.close()
        self.assertFalse(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 5)
        cursor.execute('select test')
        self.assertTrue(db._con._con.valid)
        self.assertEqual(Connection.num_pings, 6)
        Connection.has_ping = False
        Connection.num_pings = 0

    def test_failed_transaction(self):
        dbapi.threadsafety = 2
        pool = PooledDB(dbapi, 0, 1, 1)
        db = pool.connection()
        cursor = db.cursor()
        db._con._con.close()
        cursor.execute('select test')
        db.begin()
        db._con._con.close()
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test')
        cursor.execute('select test')
        db.begin()
        db.cancel()
        db._con._con.close()
        cursor.execute('select test')
        pool = PooledDB(dbapi, 1, 1, 0)
        db = pool.connection()
        cursor = db.cursor()
        db._con._con.close()
        cursor.execute('select test')
        db.begin()
        db._con._con.close()
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test')
        cursor.execute('select test')
        db.begin()
        db.cancel()
        db._con._con.close()
        cursor.execute('select test')

    def test_shared_in_transaction(self):
        dbapi.threadsafety = 2
        pool = PooledDB(dbapi, 0, 1, 1)
        db = pool.connection()
        db.begin()
        pool.connection(False)
        self.assertRaises(TooManyConnections, pool.connection)
        pool = PooledDB(dbapi, 0, 2, 2)
        db1 = pool.connection()
        db2 = pool.connection()
        self.assertIsNot(db2._con, db1._con)
        db2.close()
        db2 = pool.connection()
        self.assertIsNot(db2._con, db1._con)
        db = pool.connection()
        self.assertIs(db._con, db1._con)
        db.close()
        db1.begin()
        db = pool.connection()
        self.assertIs(db._con, db2._con)
        db.close()
        db2.begin()
        pool.connection(False)
        self.assertRaises(TooManyConnections, pool.connection)
        db1.rollback()
        db = pool.connection()
        self.assertIs(db._con, db1._con)

    def test_reset_transaction(self):
        pool = PooledDB(dbapi, 1, 1, 0)
        db = pool.connection()
        db.begin()
        con = db._con
        self.assertTrue(con._transaction)
        self.assertEqual(con._con.session, ['rollback'])
        db.close()
        self.assertIs(pool.connection()._con, con)
        self.assertFalse(con._transaction)
        self.assertEqual(con._con.session, ['rollback'] * 3)
        pool = PooledDB(dbapi, 1, 1, 0, reset=False)
        db = pool.connection()
        db.begin()
        con = db._con
        self.assertTrue(con._transaction)
        self.assertEqual(con._con.session, [])
        db.close()
        self.assertIs(pool.connection()._con, con)
        self.assertFalse(con._transaction)
        self.assertEqual(con._con.session, ['rollback'])

    def test_context_manager(self):
        pool = PooledDB(dbapi, 1, 1, 1)
        con = pool._idle_cache[0]._con
        with pool.connection() as db:
            self.assertTrue(hasattr(db, '_shared_con'))
            self.assertFalse(pool._idle_cache)
            self.assertTrue(con.valid)
            with db.cursor() as cursor:
                self.assertEqual(con.open_cursors, 1)
                cursor.execute('select test')
                r = cursor.fetchone()
            self.assertEqual(con.open_cursors, 0)
            self.assertEqual(r, 'test')
            self.assertEqual(con.num_queries, 1)
        self.assertTrue(pool._idle_cache)
        with pool.dedicated_connection() as db:
            self.assertFalse(hasattr(db, '_shared_con'))
            self.assertFalse(pool._idle_cache)
            with db.cursor() as cursor:
                self.assertEqual(con.open_cursors, 1)
                cursor.execute('select test')
                r = cursor.fetchone()
            self.assertEqual(con.open_cursors, 0)
            self.assertEqual(r, 'test')
            self.assertEqual(con.num_queries, 2)
        self.assertTrue(pool._idle_cache)


class TestSharedDBConnection(unittest.TestCase):

    def test_create_connection(self):
        db_con = dbapi.connect()
        con = SharedDBConnection(db_con)
        self.assertEqual(con.con, db_con)
        self.assertEqual(con.shared, 1)

    def test_share_and_unshare(self):
        con = SharedDBConnection(dbapi.connect())
        self.assertEqual(con.shared, 1)
        con.share()
        self.assertEqual(con.shared, 2)
        con.share()
        self.assertEqual(con.shared, 3)
        con.unshare()
        self.assertEqual(con.shared, 2)
        con.unshare()
        self.assertEqual(con.shared, 1)

    def test_comparison(self):
        con1 = SharedDBConnection(dbapi.connect())
        con1.con._transaction = False
        con2 = SharedDBConnection(dbapi.connect())
        con2.con._transaction = False
        self.assertTrue(con1 == con2)
        self.assertTrue(con1 <= con2)
        self.assertTrue(con1 >= con2)
        self.assertFalse(con1 != con2)
        self.assertFalse(con1 < con2)
        self.assertFalse(con1 > con2)
        con2.share()
        self.assertFalse(con1 == con2)
        self.assertTrue(con1 <= con2)
        self.assertFalse(con1 >= con2)
        self.assertTrue(con1 != con2)
        self.assertTrue(con1 < con2)
        self.assertFalse(con1 > con2)
        con1.con._transaction = True
        self.assertFalse(con1 == con2)
        self.assertFalse(con1 <= con2)
        self.assertTrue(con1 >= con2)
        self.assertTrue(con1 != con2)
        self.assertFalse(con1 < con2)
        self.assertTrue(con1 > con2)


if __name__ == '__main__':
    unittest.main()



"""Test the PersistentDB module.

Note:
We don't test performance here, so the test does not predicate
whether PersistentDB actually will help in improving performance or not.
We also assume that the underlying SteadyDB connections are tested.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest

from . import mock_db as dbapi

from dbutils.persistent_db import PersistentDB, local


class TestPersistentDB(unittest.TestCase):

    def setUp(self):
        dbapi.threadsafety = 1

    def test_version(self):
        from dbutils import __version__, persistent_db
        self.assertEqual(persistent_db.__version__, __version__)
        self.assertEqual(PersistentDB.version, __version__)

    def test_no_threadsafety(self):
        from dbutils.persistent_db import NotSupportedError
        for dbapi.threadsafety in (None, 0):
            self.assertRaises(NotSupportedError, PersistentDB, dbapi)

    def test_close(self):
        for closeable in (False, True):
            persist = PersistentDB(dbapi, closeable=closeable)
            db = persist.connection()
            self.assertTrue(db._con.valid)
            db.close()
            self.assertTrue(closeable ^ db._con.valid)
            db.close()
            self.assertTrue(closeable ^ db._con.valid)
            db._close()
            self.assertFalse(db._con.valid)
            db._close()
            self.assertFalse(db._con.valid)

    def test_connection(self):
        persist = PersistentDB(dbapi)
        db = persist.connection()
        db_con = db._con
        self.assertIsNone(db_con.database)
        self.assertIsNone(db_con.user)
        db2 = persist.connection()
        self.assertEqual(db, db2)
        db3 = persist.dedicated_connection()
        self.assertEqual(db, db3)
        db3.close()
        db2.close()
        db.close()

    def test_threads(self):
        num_threads = 3
        persist = PersistentDB(dbapi, closeable=True)
        from queue import Queue, Empty
        query_queue, result_queue = [], []
        for i in range(num_threads):
            query_queue.append(Queue(1))
            result_queue.append(Queue(1))

        def run_queries(i):
            this_db = persist.connection()
            while 1:
                try:
                    try:
                        q = query_queue[i].get(1, 1)
                    except TypeError:
                        q = query_queue[i].get(1)
                except Empty:
                    q = None
                if not q:
                    break
                db = persist.connection()
                if db != this_db:
                    r = 'error - not persistent'
                else:
                    if q == 'ping':
                        r = 'ok - thread alive'
                    elif q == 'close':
                        db.close()
                        r = 'ok - connection closed'
                    else:
                        cursor = db.cursor()
                        cursor.execute(q)
                        r = cursor.fetchone()
                        cursor.close()
                r = f'{i}({db._usage}): {r}'
                try:
                    result_queue[i].put(r, 1, 1)
                except TypeError:
                    result_queue[i].put(r, 1)
            db.close()

        from threading import Thread
        threads = []
        for i in range(num_threads):
            thread = Thread(target=run_queries, args=(i,))
            threads.append(thread)
            thread.start()
        for i in range(num_threads):
            try:
                query_queue[i].put('ping', 1, 1)
            except TypeError:
                query_queue[i].put('ping', 1)
        for i in range(num_threads):
            try:
                r = result_queue[i].get(1, 1)
            except TypeError:
                r = result_queue[i].get(1)
            self.assertEqual(r, f'{i}(0): ok - thread alive')
            self.assertTrue(threads[i].is_alive())
        for i in range(num_threads):
            for j in range(i + 1):
                try:
                    query_queue[i].put(f'select test{j}', 1, 1)
                    r = result_queue[i].get(1, 1)
                except TypeError:
                    query_queue[i].put(f'select test{j}', 1)
                    r = result_queue[i].get(1)
                self.assertEqual(r, f'{i}({j + 1}): test{j}')
        try:
            query_queue[1].put('select test4', 1, 1)
        except TypeError:
            query_queue[1].put('select test4', 1)
        try:
            r = result_queue[1].get(1, 1)
        except TypeError:
            r = result_queue[1].get(1)
        self.assertEqual(r, '1(3): test4')
        try:
            query_queue[1].put('close', 1, 1)
            r = result_queue[1].get(1, 1)
        except TypeError:
            query_queue[1].put('close', 1)
            r = result_queue[1].get(1)
        self.assertEqual(r, '1(3): ok - connection closed')
        for j in range(2):
            try:
                query_queue[1].put(f'select test{j}', 1, 1)
                r = result_queue[1].get(1, 1)
            except TypeError:
                query_queue[1].put(f'select test{j}', 1)
                r = result_queue[1].get(1)
            self.assertEqual(r, f'1({j + 1}): test{j}')
        for i in range(num_threads):
            self.assertTrue(threads[i].is_alive())
            try:
                query_queue[i].put('ping', 1, 1)
            except TypeError:
                query_queue[i].put('ping', 1)
        for i in range(num_threads):
            try:
                r = result_queue[i].get(1, 1)
            except TypeError:
                r = result_queue[i].get(1)
            self.assertEqual(r, f'{i}({i + 1}): ok - thread alive')
            self.assertTrue(threads[i].is_alive())
        for i in range(num_threads):
            try:
                query_queue[i].put(None, 1, 1)
            except TypeError:
                query_queue[i].put(None, 1)

    def test_maxusage(self):
        persist = PersistentDB(dbapi, 20)
        db = persist.connection()
        self.assertEqual(db._maxusage, 20)
        for i in range(100):
            cursor = db.cursor()
            cursor.execute(f'select test{i}')
            r = cursor.fetchone()
            cursor.close()
            self.assertEqual(r, f'test{i}')
            self.assertTrue(db._con.valid)
            j = i % 20 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db._con.num_uses, j)
            self.assertEqual(db._con.num_queries, j)

    def test_setsession(self):
        persist = PersistentDB(dbapi, 3, ('set datestyle',))
        db = persist.connection()
        self.assertEqual(db._maxusage, 3)
        self.assertEqual(db._setsession_sql, ('set datestyle',))
        self.assertEqual(db._con.session, ['datestyle'])
        cursor = db.cursor()
        cursor.execute('set test')
        cursor.fetchone()
        cursor.close()
        for i in range(3):
            self.assertEqual(db._con.session, ['datestyle', 'test'])
            cursor = db.cursor()
            cursor.execute('select test')
            cursor.fetchone()
            cursor.close()
        self.assertEqual(db._con.session, ['datestyle'])

    def test_threadlocal(self):
        persist = PersistentDB(dbapi)
        self.assertTrue(isinstance(persist.thread, local))

        class threadlocal:
            pass

        persist = PersistentDB(dbapi, threadlocal=threadlocal)
        self.assertTrue(isinstance(persist.thread, threadlocal))

    def test_ping_check(self):
        Connection = dbapi.Connection
        Connection.has_ping = True
        Connection.num_pings = 0
        persist = PersistentDB(dbapi, 0, None, None, 0, True)
        db = persist.connection()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        db.close()
        db = persist.connection()
        self.assertFalse(db._con.valid)
        self.assertEqual(Connection.num_pings, 0)
        persist = PersistentDB(dbapi, 0, None, None, 1, True)
        db = persist.connection()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 1)
        db.close()
        db = persist.connection()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 2)
        persist = PersistentDB(dbapi, 0, None, None, 2, True)
        db = persist.connection()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 2)
        db.close()
        db = persist.connection()
        self.assertFalse(db._con.valid)
        self.assertEqual(Connection.num_pings, 2)
        cursor = db.cursor()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        cursor.execute('select test')
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        persist = PersistentDB(dbapi, 0, None, None, 4, True)
        db = persist.connection()
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        db.close()
        db = persist.connection()
        self.assertFalse(db._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        cursor = db.cursor()
        db._con.close()
        self.assertFalse(db._con.valid)
        self.assertEqual(Connection.num_pings, 3)
        cursor.execute('select test')
        self.assertTrue(db._con.valid)
        self.assertEqual(Connection.num_pings, 4)
        Connection.has_ping = False
        Connection.num_pings = 0

    def test_failed_transaction(self):
        persist = PersistentDB(dbapi)
        db = persist.connection()
        cursor = db.cursor()
        db._con.close()
        cursor.execute('select test')
        db.begin()
        db._con.close()
        self.assertRaises(dbapi.InternalError, cursor.execute, 'select test')
        cursor.execute('select test')
        db.begin()
        db.cancel()
        db._con.close()
        cursor.execute('select test')

    def test_context_manager(self):
        persist = PersistentDB(dbapi)
        with persist.connection() as db:
            with db.cursor() as cursor:
                cursor.execute('select test')
                r = cursor.fetchone()
            self.assertEqual(r, 'test')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3.9

"""Build HTML from reST files."""

from glob import glob
from os.path import splitext
from docutils.core import publish_file

print("Creating the documentation...")

for rst_file in glob('*.rst'):
    name = splitext(rst_file)[0]
    lang = splitext(name)[1]
    if lang.startswith('.'):
        lang = lang[1:]
        if lang == 'zh':
            lang = 'zh_cn'
    else:
        lang = 'en'
    html_file = name + '.html'
    print(name, lang)

    with open(rst_file, encoding='utf-8-sig') as source:
        with open(html_file, 'w', encoding='utf-8') as destination:
            output = publish_file(
                writer_name='html5', source=source, destination=destination,
                enable_exit_status=True,
                settings_overrides=dict(
                    stylesheet_path='doc.css',
                    embed_stylesheet=False,
                    toc_backlinks=False,
                    language_code=lang,
                    exit_status_level=2))

print("Done.")



"""Test the SteadyPg module.

Note:
We do not test the real PyGreSQL module, but we just
mock the basic connection functionality of that module.
We assume that the PyGreSQL module will detect lost
connections correctly and set the status flag accordingly.

Copyright and credit info:

* This test was contributed by Christoph Zwerschke
"""

import unittest
import sys

from . import mock_pg as pg

from dbutils.steady_pg import SteadyPgConnection


class TestSteadyPg(unittest.TestCase):

    def test_version(self):
        from dbutils import __version__, steady_pg
        self.assertEqual(steady_pg.__version__, __version__)
        self.assertEqual(steady_pg.SteadyPgConnection.version, __version__)

    def test_mocked_connection(self):
        PgConnection = pg.DB
        db = PgConnection(
            'SteadyPgTestDB', user='SteadyPgTestUser')
        self.assertTrue(hasattr(db, 'db'))
        self.assertTrue(hasattr(db.db, 'status'))
        self.assertTrue(db.db.status)
        self.assertTrue(hasattr(db.db, 'query'))
        self.assertTrue(hasattr(db.db, 'close'))
        self.assertFalse(hasattr(db.db, 'reopen'))
        self.assertTrue(hasattr(db, 'reset'))
        self.assertTrue(hasattr(db.db, 'num_queries'))
        self.assertTrue(hasattr(db.db, 'session'))
        self.assertFalse(hasattr(db.db, 'get_tables'))
        self.assertTrue(hasattr(db.db, 'db'))
        self.assertEqual(db.db.db, 'SteadyPgTestDB')
        self.assertTrue(hasattr(db.db, 'user'))
        self.assertEqual(db.db.user, 'SteadyPgTestUser')
        self.assertTrue(hasattr(db, 'query'))
        self.assertTrue(hasattr(db, 'close'))
        self.assertTrue(hasattr(db, 'reopen'))
        self.assertTrue(hasattr(db, 'reset'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertTrue(hasattr(db, 'session'))
        self.assertTrue(hasattr(db, 'get_tables'))
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'SteadyPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'SteadyPgTestUser')
        for i in range(3):
            self.assertEqual(db.num_queries, i)
            self.assertEqual(
                db.query(f'select test{i}'), f'test{i}')
        self.assertTrue(db.db.status)
        db.reopen()
        self.assertTrue(db.db.status)
        self.assertEqual(db.num_queries, 0)
        self.assertEqual(db.query('select test4'), 'test4')
        self.assertEqual(db.get_tables(), 'test')
        db.close()
        try:
            status = db.db.status
        except AttributeError:
            status = False
        self.assertFalse(status)
        self.assertRaises(pg.InternalError, db.close)
        self.assertRaises(pg.InternalError, db.query, 'select test')
        self.assertRaises(pg.InternalError, db.get_tables)

    def test_broken_connection(self):
        self.assertRaises(TypeError, SteadyPgConnection, 'wrong')
        db = SteadyPgConnection(dbname='ok')
        InternalError = sys.modules[db._con.__module__].InternalError
        for i in range(3):
            db.close()
        del db
        self.assertRaises(InternalError, SteadyPgConnection, dbname='error')

    def test_close(self):
        for closeable in (False, True):
            db = SteadyPgConnection(closeable=closeable)
            self.assertTrue(db._con.db and db._con.valid)
            db.close()
            self.assertTrue(
                closeable ^ (db._con.db is not None and db._con.valid))
            db.close()
            self.assertTrue(
                closeable ^ (db._con.db is not None and db._con.valid))
            db._close()
            self.assertFalse(db._con.db and db._con.valid)
            db._close()
            self.assertFalse(db._con.db and db._con.valid)

    def test_connection(self):
        db = SteadyPgConnection(
            0, None, 1, 'SteadyPgTestDB', user='SteadyPgTestUser')
        self.assertTrue(hasattr(db, 'db'))
        self.assertTrue(hasattr(db, '_con'))
        self.assertEqual(db.db, db._con.db)
        self.assertTrue(hasattr(db, '_usage'))
        self.assertEqual(db._usage, 0)
        self.assertTrue(hasattr(db.db, 'status'))
        self.assertTrue(db.db.status)
        self.assertTrue(hasattr(db.db, 'query'))
        self.assertTrue(hasattr(db.db, 'close'))
        self.assertFalse(hasattr(db.db, 'reopen'))
        self.assertTrue(hasattr(db.db, 'reset'))
        self.assertTrue(hasattr(db.db, 'num_queries'))
        self.assertTrue(hasattr(db.db, 'session'))
        self.assertTrue(hasattr(db.db, 'db'))
        self.assertEqual(db.db.db, 'SteadyPgTestDB')
        self.assertTrue(hasattr(db.db, 'user'))
        self.assertEqual(db.db.user, 'SteadyPgTestUser')
        self.assertFalse(hasattr(db.db, 'get_tables'))
        self.assertTrue(hasattr(db, 'query'))
        self.assertTrue(hasattr(db, 'close'))
        self.assertTrue(hasattr(db, 'reopen'))
        self.assertTrue(hasattr(db, 'reset'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertTrue(hasattr(db, 'session'))
        self.assertTrue(hasattr(db, 'dbname'))
        self.assertEqual(db.dbname, 'SteadyPgTestDB')
        self.assertTrue(hasattr(db, 'user'))
        self.assertEqual(db.user, 'SteadyPgTestUser')
        self.assertTrue(hasattr(db, 'get_tables'))
        for i in range(3):
            self.assertEqual(db._usage, i)
            self.assertEqual(db.num_queries, i)
            self.assertEqual(
                db.query(f'select test{i}'), f'test{i}')
        self.assertTrue(db.db.status)
        self.assertEqual(db.get_tables(), 'test')
        self.assertTrue(db.db.status)
        self.assertEqual(db._usage, 4)
        self.assertEqual(db.num_queries, 3)
        db.reopen()
        self.assertTrue(db.db.status)
        self.assertEqual(db._usage, 0)
        self.assertEqual(db.num_queries, 0)
        self.assertEqual(db.query('select test'), 'test')
        self.assertTrue(db.db.status)
        self.assertTrue(hasattr(db._con, 'status'))
        self.assertTrue(db._con.status)
        self.assertTrue(hasattr(db._con, 'close'))
        self.assertTrue(hasattr(db._con, 'query'))
        db.close()
        try:
            status = db.db.status
        except AttributeError:
            status = False
        self.assertFalse(status)
        self.assertTrue(hasattr(db._con, 'close'))
        self.assertTrue(hasattr(db._con, 'query'))
        InternalError = sys.modules[db._con.__module__].InternalError
        self.assertRaises(InternalError, db._con.close)
        self.assertRaises(InternalError, db._con.query, 'select test')
        self.assertEqual(db.query('select test'), 'test')
        self.assertTrue(db.db.status)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db.num_queries, 1)
        db.db.status = False
        self.assertFalse(db.db.status)
        self.assertEqual(db.query('select test'), 'test')
        self.assertTrue(db.db.status)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db.num_queries, 1)
        db.db.status = False
        self.assertFalse(db.db.status)
        self.assertEqual(db.get_tables(), 'test')
        self.assertTrue(db.db.status)
        self.assertEqual(db._usage, 1)
        self.assertEqual(db.num_queries, 0)

    def test_connection_context_handler(self):
        db = SteadyPgConnection(
            0, None, 1, 'SteadyPgTestDB', user='SteadyPgTestUser')
        self.assertEqual(db.session, [])
        with db:
            db.query('select test')
        self.assertEqual(db.session, ['begin', 'commit'])
        try:
            with db:
                db.query('error')
        except pg.ProgrammingError:
            error = True
        else:
            error = False
        self.assertTrue(error)
        self.assertEqual(
            db._con.session, ['begin', 'commit', 'begin', 'rollback'])

    def test_connection_maxusage(self):
        db = SteadyPgConnection(10)
        for i in range(100):
            r = db.query(f'select test{i}')
            self.assertEqual(r, f'test{i}')
            self.assertTrue(db.db.status)
            j = i % 10 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db.num_queries, j)
        db.begin()
        for i in range(100):
            r = db.get_tables()
            self.assertEqual(r, 'test')
            self.assertTrue(db.db.status)
            if i == 49:
                db.commit()
            j = i % 10 + 1 if i > 49 else i + 11
            self.assertEqual(db._usage, j)
            j = 0 if i > 49 else 10
            self.assertEqual(db.num_queries, j)
        for i in range(10):
            if i == 7:
                db.db.status = False
            r = db.query(f'select test{i}')
            self.assertEqual(r, f'test{i}')
            j = i % 7 + 1
            self.assertEqual(db._usage, j)
            self.assertEqual(db.num_queries, j)
        for i in range(10):
            if i == 5:
                db.db.status = False
            r = db.get_tables()
            self.assertEqual(r, 'test')
            j = (i + (3 if i < 5 else -5)) % 10 + 1
            self.assertEqual(db._usage, j)
            j = 3 if i < 5 else 0
            self.assertEqual(db.num_queries, j)
        db.close()
        self.assertEqual(db.query('select test1'), 'test1')
        self.assertEqual(db._usage, 1)
        self.assertEqual(db.num_queries, 1)
        db.reopen()
        self.assertEqual(db._usage, 0)
        self.assertEqual(db.num_queries, 0)
        self.assertEqual(db.query('select test2'), 'test2')
        self.assertEqual(db._usage, 1)
        self.assertEqual(db.num_queries, 1)

    def test_connection_setsession(self):
        db = SteadyPgConnection(3, ('set time zone', 'set datestyle'))
        self.assertTrue(hasattr(db, 'num_queries'))
        self.assertEqual(db.num_queries, 0)
        self.assertTrue(hasattr(db, 'session'))
        self.assertEqual(tuple(db.session), ('time zone', 'datestyle'))
        for i in range(11):
            db.query('select test')
        self.assertEqual(db.num_queries, 2)
        self.assertEqual(db.session, ['time zone', 'datestyle'])
        db.query('set test')
        self.assertEqual(db.num_queries, 2)
        self.assertEqual(db.session, ['time zone', 'datestyle', 'test'])
        db.query('select test')
        self.assertEqual(db.num_queries, 1)
        self.assertEqual(db.session, ['time zone', 'datestyle'])
        db.close()
        db.query('set test')
        self.assertEqual(db.num_queries, 0)
        self.assertEqual(db.session, ['time zone', 'datestyle', 'test'])

    def test_begin(self):
        for closeable in (False, True):
            db = SteadyPgConnection(closeable=closeable)
            db.begin()
            self.assertEqual(db.session, ['begin'])
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            db.close()
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            db.begin()
            self.assertEqual(db.session, ['begin'])
            db.db.close()
            self.assertRaises(pg.InternalError, db.query, 'select test')
            self.assertEqual(db.num_queries, 0)
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            self.assertEqual(db.begin('select sql:begin'), 'sql:begin')
            self.assertEqual(db.num_queries, 2)

    def test_end(self):
        for closeable in (False, True):
            db = SteadyPgConnection(closeable=closeable)
            db.begin()
            db.query('select test')
            db.end()
            self.assertEqual(db.session, ['begin', 'end'])
            db.db.close()
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            self.assertEqual(db.begin('select sql:end'), 'sql:end')
            self.assertEqual(db.num_queries, 2)
            db.begin()
            db.query('select test')
            db.commit()
            self.assertEqual(db.session, ['begin', 'commit'])
            db.db.close()
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            self.assertEqual(db.begin('select sql:commit'), 'sql:commit')
            self.assertEqual(db.num_queries, 2)
            db.begin()
            db.query('select test')
            db.rollback()
            self.assertEqual(db.session, ['begin', 'rollback'])
            db.db.close()
            db.query('select test')
            self.assertEqual(db.num_queries, 1)
            self.assertEqual(db.begin('select sql:rollback'), 'sql:rollback')
            self.assertEqual(db.num_queries, 2)


if __name__ == '__main__':
    unittest.main()



"""Setup Script for DBUtils"""

import warnings
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from sys import version_info

py_version = version_info[:2]
if not (3, 6) <= py_version < (4, 0):
    raise ImportError(
        'Python {}.{} is not supported by DBUtils.'.format(*py_version))

warnings.filterwarnings('ignore', 'Unknown distribution option')

__version__ = '3.0.3'

readme = open('README.md').read()

setup(
    name='DBUtils',
    version=__version__,
    description='Database connections for multi-threaded environments.',
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
    ],
    author='Christoph Zwerschke',
    author_email='cito@online.de',
    url='https://webwareforpython.github.io/DBUtils/',
    download_url="https://pypi.org/project/DBUtils/",
    project_urls={
        'Documentation':
            'https://webwareforpython.github.io/DBUtils/main.html',
        'Changelog':
            'https://webwareforpython.github.io/DBUtils/changelog.html',
        'Issue Tracker':
            'https://github.com/WebwareForPython/DBUtils/issues',
        'Source Code':
            'https://github.com/WebwareForPython/DBUtils'},
    platforms=['any'],
    license='MIT License',
    packages=['dbutils']
)
